from .priors import *
from .pre_grid import *
from .gp_sfh import *
from .plotter import *
from .sed_fitter import *


def sample_fesc_prior():
    # uniform between 0 and 1
    return np.random.uniform()


def get_k(f_esc = 0.0, f_dust = 0.0):    
    # factor for downscaling nebular line spectrum, from Inoue (2011)
    # taken from Boquien+20: https://www.aanda.org/articles/aa/full_html/2019/02/aa34156-18/aa34156-18.html
    
    # alphaB_Te: case B recombination rate, in m^3/s
    # alpha1_Te: alpha_A-alpha_B - recombination rate to the ground level
    # Te: electron temp in kelvin
    # fesc: lyman continuum escape fraction
    # fdust: partial absorption by dust before ionization
    
    Te = 1e4 # kelvin
    alphaB_Te = 2.58e-19 #m^3/s
    alpha1_Te = 1.54e-19 #m^3/s
    
    alpharatio = alpha1_Te/alphaB_Te
    
    k = (1-f_esc - f_dust) / (1+(alpharatio)*(f_esc+f_dust))
    
    return k


def makespec_fesc(specdetails, fesc, priors, sp, cosmo, filter_list = [], filt_dir = [], return_spec = False, peraa = False, input_sfh = False):
    """
    makespec() works in two ways to create spectra or SEDs from an input list of physical paramters. 
    with input_sfh = False, specdetails = [sfh_tuple, dust, met, zval]
    with input_sfh = True, specdetails = [sfh, timeax, dust, met, zval]
    
    return_spec can be true, false, or an array of wavelengths. in case 
    
    it uses the db.mocksp object for the underlying SPS calculations, so make sure it's set accordingly.
    it also uses the priors object for things like decouple_sfr. 
    
    """
    
    
    # hardcoded parameters - offload this to a seprarate function
    sp.params['sfh'] = 3
    sp.params['cloudy_dust'] = True
    sp.params['gas_logu'] = -2
    sp.params['add_igm_absorption'] = True
    sp.params['add_neb_emission'] = True
    sp.params['add_neb_continuum'] = True
    sp.params['imf_type'] = 1 # Chabrier
    
    # variable parameters
    if input_sfh == True:
        [sfh, timeax, dust, met, zval] = specdetails
    else:
        [sfh_tuple, dust, met, zval] = specdetails
        sfh, timeax = tuple_to_sfh(sfh_tuple, zval, decouple_sfr = priors.decouple_sfr, decouple_sfr_time = priors.decouple_sfr_time)
    sp.set_tabular_sfh(timeax, sfh)
    # sp.params['dust_type'] = 2
    # sp.params['dust1'] = dust1_rand
    sp.params['dust2'] = dust
    sp.params['logzsol'] = met
    sp.params['gas_logz'] = met # matching stellar to gas-phase metallicity
    sp.params['zred'] = zval
    
    #lam, spec = sp.get_spectrum(tage = cosmo.age(zval).value, peraa = peraa)
    # adding 0.1 Myr here to get the last couple of FSPS SSPs
    
    sp.params['add_neb_emission'] = False
    sp.params['add_neb_continuum'] = False
    lam, spec_noneb_dusty = sp.get_spectrum(tage = cosmo.age(zval).value+1e-4, peraa = peraa)
    
    sp.params['add_neb_emission'] = True
    sp.params['add_neb_continuum'] = True
    lam, spec_dusty = sp.get_spectrum(tage = cosmo.age(zval).value+1e-4, peraa = peraa)

    neb_emission_dusty = spec_dusty - spec_noneb_dusty
    k_factor = get_k(f_esc = fesc)
    neb_emission_dusty = neb_emission_dusty * k_factor
    spec = spec_noneb_dusty + neb_emission_dusty
    
#     lam, spec = sp.get_spectrum(tage = cosmo.age(zval).value+1e-4, peraa = peraa)
    spec_ujy = convert_to_microjansky(spec, zval, cosmo)
    
    if type(return_spec) == type(True):
        
        if return_spec == True:
            return lam, spec_ujy
    
        elif return_spec == False:
            filcurves, _, _ = make_filvalkit_simple(lam, zval, fkit_name = filter_list, filt_dir = filt_dir)
            sed = calc_fnu_sed_fast(spec_ujy, filcurves)
            return sed
    
    elif len(return_spec) > 10:
        return convert_to_splined_spec(spec, lam, return_spec, zval)
    
    else:
        raise('Unknown argument for return_spec. Use True or False, or pass a wavelength grid.')
        
    return 0


def generate_atlas_fesc(N_pregrid = 10, priors=priors, initial_seed = 42, store = True, filter_list = 'filter_list.dat', filt_dir = 'filters/', norm_method = 'median', z_step = 0.01, sp = mocksp, cosmology = cosmo, fname = None, path = 'pregrids/', lam_array_spline = [], rseed = None):

    """Generate a pregrid of galaxy properties and their corresponding SEDs
        drawn from the prior distributions defined in priors.py
        Args:
            N_pregrid[int]: Number of SEDs in the pre-grid.
            Nparam[int]: Number of SFH parameters in each tuple
            Initial_seed[int]: Initial seed for random number generation.
            store[Boolean, optional]: Flag whether to store results
                or return as output
            filter_list[filename]: File that contains a list of filter curves.
            z_step[float]: Step size in redshift for filter curve grid.
                Default is 0.01.
            norm_method[string, default = 'max']: normalization for SEDs and SFHs.
                Currently supported arguments are 'none', 'max', 'median', 'area'.
            sp[stellar population object]: FSPS stellar population object. Initialized previously for speed.
            cosmo[astropy cosmology object]: cosmology. Default is FlatLambdaCDM
        Returns:
            [if store == False]
            rand_sfh_tuples[2d numpy array]: N_samples prior-sampled SFH tuples
            rand_Z: prior-sampled metallicity values
            rand_Av: prior-sampled dust attenuation values
            rand_z: prior-sampled redshift values
            rand_seds: Corresponding SEDs in F_\nu (\muJy)
            norm_method: Argument for how SEDs are normalized, pass into fitter
        """

    print('generating atlas with: ')
    print(priors.Nparam, ' tx parameters, ', priors.sfr_prior_type, ' SFR sampling', priors.sfh_treatment,' SFH treatment', priors.met_treatment,' met sampling', priors.dust_model, ' dust attenuation', priors.dust_prior,' dust prior', priors.decouple_sfr,' SFR decoupling.')
    
    if rseed is not None:
        print('setting random seed to :',rseed)
        np.random.seed(rseed)

    zval_all = []
    sfh_tuple_all = []
    sfh_tuple_rec_all = []
    norm_all = []
    dust_all = []
    met_all = []
    sed_all = []
    mstar_all = []
    sfr_all = []
    sim_timeax_all = []
    sim_sfh_all = []
    
    fesc_all = []

    Nparam = priors.Nparam

    for i in tqdm(range(int(N_pregrid))):

        zval = priors.sample_z_prior()

        massval = priors.sample_mass_prior()
        if priors.sfr_prior_type == 'sSFRlognormal':
            sfrval = priors.sample_sfr_prior(zval=zval)
        else:
            sfrval = priors.sample_sfr_prior()
        txparam = priors.sample_tx_prior()
        sfh_tuple = np.hstack((massval, sfrval, Nparam, txparam))
        norm = 1.0
        
        if priors.dynamic_decouple == True:
            priors.decouple_sfr_time = 100*cosmo.age(zval).value/cosmo.age(0.1).value
            #print('decouple time: %.2f myr at z: %.2f' %(priors.decouple_sfr_time,zval))
        sfh, timeax = tuple_to_sfh(sfh_tuple, zval, decouple_sfr = priors.decouple_sfr, decouple_sfr_time = priors.decouple_sfr_time)
        
        temp = calctimes(timeax, sfh, priors.Nparam)
        temptuple = calctimes_to_tuple(temp)
        
        dust = priors.sample_Av_prior()
        met = priors.sample_Z_prior()

        specdetails = [sfh_tuple, dust, met, zval]
        
        try:
            fesc = priors.sample_fesc_prior() # uniform between 0 & 1
        except:
            raise Exception('Fesc prior undefined. Define a priors.sample_fesc_prior()')
    
        if len(lam_array_spline) > 0:
            sed = makespec_fesc(specdetails, fesc, priors, sp, cosmology, filter_list, filt_dir, return_spec = lam_array_spline, peraa = True)
        else:
            lam, spec_ujy = makespec_fesc(specdetails, fesc, priors, sp, cosmology, filter_list, filt_dir, return_spec = True)

            if i == 0:
                # make grid of filter transmission curves for faster computation
                fc_zgrid = np.arange(priors.z_min-z_step, priors.z_max+z_step, z_step)
                temp_fc, temp_lz, temp_lz_lores = make_filvalkit_simple(lam,priors.z_min,fkit_name = filter_list, filt_dir = filt_dir)

                fcs= np.zeros((temp_fc.shape[0], temp_fc.shape[1], len(fc_zgrid)))
                lzs = np.zeros((temp_lz.shape[0], len(fc_zgrid)))
                lzs_lores = np.zeros((temp_lz_lores.shape[0], len(fc_zgrid)))

                for i in (range(len(fc_zgrid))):
                    fcs[0:,0:,i], lzs[0:,i], lzs_lores[0:,i] = make_filvalkit_simple(lam,fc_zgrid[i],fkit_name = filter_list, filt_dir = filt_dir)

            fc_index = np.argmin(np.abs(zval - fc_zgrid))
            sed = calc_fnu_sed_fast(spec_ujy, fcs[0:,0:,fc_index])

        #-------------------------------------------

        if norm_method == 'none':
            # no normalization
            norm_fac = 1
        elif norm_method == 'max':
            # normalize SEDs to 1 - seems to work better than median for small grids
            norm_fac = np.amax(sed)
        elif norm_method == 'median':
            # normalize SEDs to median
            norm_fac = np.nanmedian(sed)
        elif norm_method == 'area':
            # normalize SFH to 10^9 Msun
            norm_fac == 10**(massval - 9)
        else:
            raise ValueError('undefined normalization argument')

        sed = sed/norm_fac
        norm = norm/norm_fac
        mstar = np.log10(sp.stellar_mass / norm_fac)
        sfr = np.log10(sp.sfr / norm_fac)
        sfh_tuple[0] = sfh_tuple[0] - np.log10(norm_fac)
        sfh_tuple[1] = sfh_tuple[1] - np.log10(norm_fac)
        temptuple[0] = temptuple[0] - np.log10(norm_fac)
        temptuple[1] = temptuple[1] - np.log10(norm_fac)

        #-------------------------------------------

        zval_all.append(zval)
        sfh_tuple_all.append(sfh_tuple)
        sfh_tuple_rec_all.append(temptuple)
        norm_all.append(norm)
        dust_all.append(dust)
        met_all.append(met)
        sed_all.append(sed)
        mstar_all.append(mstar)
        sfr_all.append(sfr)
        fesc_all.append(fesc)


    pregrid_dict = {'zval':np.array(zval_all),
                   'sfh_tuple':np.array(sfh_tuple_all),
                   'sfh_tuple_rec':np.array(sfh_tuple_rec_all),
                   'norm':np.array(norm_all), 'norm_method':norm_method,
                   'mstar':np.array(mstar_all), 'sfr':np.array(sfr_all),
                   'dust':np.array(dust_all), 'met':np.array(met_all),
                   'fesc':np.array(fesc_all),
                   'sed':np.array(sed_all)}

    if store == True:

        if fname is None:
            fname = 'sfh_pregrid_size'
        if os.path.exists(path):
            print('Path exists. Saved atlas at : '+path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas')
        else:
            os.mkdir(path)
            print('Created directory and saved atlas at : '+path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas')
        try:
            hickle.dump(pregrid_dict,
                        path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas',
                        compression='gzip', compression_opts = 9)
        except:
            print('storing without compression')
            hickle.dump(pregrid_dict,
                        path+fname+'_'+str(N_pregrid)+'_Nparam_'+str(Nparam)+'.dbatlas')
            
        return

    return pregrid_dict


def plot_posteriors_fesc(sedfit,truths = [], **kwargs):
    
    chi2_array = sedfit.chi2_array
    norm_fac = sedfit.norm_fac
    sed = sedfit.sed
    atlas = sedfit.atlas
    
    set_plot_style()

    if len(truths) > 0:
        corner_truths = truths

    sfrvals = atlas['sfr'].copy()
    sfrvals[sfrvals<-3] = -3
    pg_params = np.vstack([atlas['mstar'],
                           sfrvals,
                           atlas['sfh_tuple'][0:,3:].T,
                           atlas['met'].ravel(),
                           atlas['dust'].ravel(),
                           atlas['fesc'].ravel(),
                           atlas['zval'].ravel()])
    txs = ['t'+'%.0f' %i for i in quantile_names(pg_params.shape[0]-6)]
    pg_labels = ['log M$_*$', 'log SFR', 'Z', 'Av', 'f$_{esc}$', 'z']
    pg_labels[2:2] = txs

    corner_params = pg_params.copy()
    corner_params[0,0:] += np.log10(norm_fac)
    corner_params[1,0:] += np.log10(norm_fac)
    
    if len(truths) > 0:
        figure = corner.corner(corner_params.T, weights = np.exp(-chi2_array/2),
                                labels=pg_labels, truths=corner_truths,
                                plot_datapoints=False, fill_contours=True,
                                bins=20, smooth=1.0,
                                quantiles=(0.16, 0.84), levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
                                label_kwargs={"fontsize": 30}, show_titles=True, **kwargs)
    else:
        figure = corner.corner(corner_params.T, weights = np.exp(-chi2_array/2),
                                labels=pg_labels,
                                plot_datapoints=False, fill_contours=True,
                                bins=20, smooth=1.0,
                                quantiles=(0.16, 0.84), levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
                                label_kwargs={"fontsize": 30}, show_titles=True, **kwargs)
    figure.subplots_adjust(right=1.5,top=1.5)

    return figure          

def calc_fesc_posteriors(sedfit):
    """
    add f_esc to the calculated posteriors
    """
    fesc_vals = get_quants_key('fesc', 50, sedfit.chi2_array, sedfit.atlas, sedfit.norm_fac)
    sedfit.fesc = fesc_vals

    return
    



import numpy as np

from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF,WhiteKernel, ConstantKernel as C, DotProduct, RationalQuadratic, Matern
import george
from george import kernels
from scipy.optimize import minimize
import scipy.io as sio

import warnings
warnings.filterwarnings('ignore')

from .priors import *

fsps_mlc = sio.loadmat('dense_basis/train_data/fsps_mass_loss_curve.mat')
fsps_time = fsps_mlc['timeax_fsps'].ravel()
fsps_massloss = fsps_mlc['mass_loss_fsps'].ravel()

#-----------------------------------------------------------------------
#           Scikit-learn implementation of the GP-SFH module
#-----------------------------------------------------------------------

def gp_sfh_sklearn(sfh_tuple,zval=5.606,vb = False,std=False):

    """
    Function to generate a (possibly random?) SFH given the following input parameters:
    mass = stellar mass
    sfr = current sfr, averaged over some timescale the SED is sensitive to, 10-100Myr
    Nparam = number of parameters. 1=t50, 2=t33,t67, 3=t25,t50,75 etc
    param_arr = the numpy array of actual parameters

    """

    mass = sfh_tuple[0]
    #sfr = 10**sfh_tuple[1]
    sfr = sfh_tuple[1]
    Nparam = int(sfh_tuple[2])
    param_arr = sfh_tuple[3:]

    # fix SFH(t=0) = 0
    # fix SFH(t=tobs) = SFR
    # fit \int SFH(t)dt = mass


    #zval = 1.0

    # this is the times at which the galaxy formed x percentiles of its observed stellar mass...
    input_Nparams = Nparam
    input_params = param_arr

    # these are time quantities (tx)
    input_params_full = np.zeros((input_Nparams+5,))
    input_params_full[0] = 0
    input_params_full[1] = 0.01
    input_params_full[2:-3] = input_params
    input_params_full[-3] = 0.998
    input_params_full[-2] = 0.999
    input_params_full[-1] = 1

    # these are galaxy mass quantities M(t)
    temp_mass = np.linspace(0,1,input_Nparams+2)
    input_mass = np.zeros((input_params_full.shape))

    input_mass[1] = 0.0
    input_mass[2] = 1e-1
    input_mass[2:-3] = temp_mass[1:-1]
    input_mass[-3] = 1.0 - np.power(10,sfr)*(1.0-0.998)*(cosmo.age(zval).value*1e9)/np.power(10,mass)
    input_mass[-2] = 1.0 - np.power(10,sfr)*(1.0-0.999)*(cosmo.age(zval).value*1e9)/np.power(10,mass)
    input_mass[-1] = 1.0
    # the last two statements help to fix the SFR at t_obs

    xax = input_params_full.reshape(-1,1)
    yax = (input_mass-input_params_full).reshape(-1,1)

    #------------------------------------

    kernel = DotProduct(10.0, (1e-2,1e2)) *RationalQuadratic(0.1)
    gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=9)
    gp.fit(xax,yax)

    x = np.linspace(0,1,1000)
    y_pred, sigma = gp.predict(x[:,np.newaxis], return_std=True)
    y_pred = y_pred.ravel() + x

    mass_unnormed = y_pred*np.power(10,mass)
    time_unnormed = x*cosmo.age(zval).value*1e9

    sfr_unnormed = np.diff(mass_unnormed)/np.diff(time_unnormed)
    gen_sfh = np.zeros((mass_unnormed.shape))
    gen_sfh[1:] = sfr_unnormed

    mask = (gen_sfh<0)
    gen_sfh[mask] = 0

    #y_mean, y_cov = gp.predict(X_[:,np.newaxis], return_cov=True)
    if vb == True:

        plt.plot(input_params_full,input_mass,'k.',markersize=15)
        plt.plot(x,y_pred,'g--')
        plt.fill_between(x,y_pred.ravel()- sigma,y_pred.ravel()+sigma,alpha=0.1,color='g')
        plt.axis([0,1,0,1])
        plt.xlabel('normalized time')
        plt.ylabel('normalized mass')
        plt.show()

        plt.plot(np.amax(time_unnormed)/1e9-time_unnormed/1e9,gen_sfh,'k-',lw=3)
        plt.xlabel('t [lookback time; Gyr]',fontsize=14)
        plt.ylabel('SFR(t) [solar masses/yr]',fontsize=14)
        plt.show()

    if std == False:
        return gen_sfh, time_unnormed
    else:
        mass_sigmaup_unnormed = (y_pred.ravel()+sigma)*np.power(10,mass)
        mass_sigmadn_unnormed = (y_pred.ravel()-sigma)*np.power(10,mass)
        sfr_sigmaup_unnormed = np.diff(mass_sigmaup_unnormed)/np.diff(time_unnormed)
        sfr_sigmadn_unnormed = np.diff(mass_sigmadn_unnormed)/np.diff(time_unnormed)
        gen_sfh_up = np.zeros((mass_unnormed.shape))
        gen_sfh_dn = np.zeros((mass_unnormed.shape))
        gen_sfh_up[1:] = sfr_sigmaup_unnormed
        gen_sfh_dn[1:] = sfr_sigmadn_unnormed

        maskup = (gen_sfh_up < 0)
        maskdn = (gen_sfh_dn < 0)

        gen_sfh_up[maskup] = 0
        gen_sfh_dn[maskdn] = 0
        return gen_sfh, gen_sfh_up, gen_sfh_dn, time_unnormed

#-----------------------------------------------------------------------
#               George implementation of the GP-SFH module
#-----------------------------------------------------------------------


def neg_ln_like(p, gp, y):
    gp.set_parameter_vector(p)
    return -gp.log_likelihood(y)

def grad_neg_ln_like(p, gp, y):
    gp.set_parameter_vector(p)
    return -gp.grad_log_likelihood(y)

def gp_sfh_george(sfh_tuple,zval=5.606,vb = False,std=False, sample_posterior = False, n_samples = 10):
    """
    Create a GP-approximation of an SFH based on the george GP package by DFM:
    mass = stellar mass
    sfr = current sfr, averaged over some timescale the SED is sensitive to, 10-100Myr
    Nparam = number of parameters. 1=t50, 2=t33,t67, 3=t25,t50,75 etc
    param_arr = the numpy array of actual parameters
    """

    mass = sfh_tuple[0]
    #sfr = 10**sfh_tuple[1]
    sfr = sfh_tuple[1]
    Nparam = int(sfh_tuple[2])
    param_arr = sfh_tuple[3:]

    # this is the times at which the galaxy formed x percentiles of its observed stellar mass...
    input_Nparams = Nparam
    input_params = param_arr

    # these are time quantities (tx)
    input_params_full = np.zeros((input_Nparams+5,))
    input_params_full[0] = 0
    input_params_full[1] = 0.01
    input_params_full[2:-3] = input_params
    input_params_full[-3] = 0.998
    input_params_full[-2] = 0.999
    input_params_full[-1] = 1

    # these are galaxy mass quantities M(t)
    temp_mass = np.linspace(0,1,input_Nparams+2)
    input_mass = np.zeros((input_params_full.shape))

    input_mass[1] = 0.0
    input_mass[2] = 1e-1
    input_mass[2:-3] = temp_mass[1:-1]
    input_mass[-3] = 1.0 - np.power(10,sfr)*(1.0-0.998)*(cosmo.age(zval).value*1e9)/np.power(10,mass)
    input_mass[-2] = 1.0 - np.power(10,sfr)*(1.0-0.999)*(cosmo.age(zval).value*1e9)/np.power(10,mass)
    input_mass[-1] = 1.0
    # the last two statements help to fix the SFR at t_obs

    xax = input_params_full
    yax = (input_mass)
    yerr = np.zeros_like(yax)
    #yerr[1:-3] = yax[1:-3]*0.05
    #yerr[2:-3] = (1-yax[2:-3])*0.05
    #yerr[2:-3] = 0.05
    yerr[2:-3] = 0.001/np.sqrt(Nparam)
    if Nparam > 20:
        yerr[2:-3] = 0.1/np.sqrt(Nparam)

    #------------------------------------

    #kernel = DotProduct(10.0, (1e-2,1e2)) *RationalQuadratic(0.1)
    #gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=9)
    #gp.fit(xax,yax)

    #x = np.linspace(0,1,1000)
    #y_pred, sigma = gp.predict(x[:,np.newaxis], return_std=True)
    #y_pred = y_pred.ravel() + x

    #------------------------------------


    #kernel = np.var(yax) * kernels.ExpSquaredKernel(np.median(yax)+np.std(yax))
    #k2 = np.var(yax) * kernels.LinearKernel(np.median(yax),order=1)
    kernel = np.var(yax) * kernels.Matern32Kernel(np.median(yax)) #+ k2
    gp = george.GP(kernel)

    #print(xax.shape, yerr.shape)

    gp.compute(xax.ravel(), yerr.ravel())

    x_pred = np.linspace(np.amin(xax), np.amax(xax), 1000)
    pred, pred_var = gp.predict(yax.ravel(), x_pred, return_var=True)
    y_pred = pred

    if sample_posterior == True:

        #print('this is happening!')
        samples = np.zeros((len(x_pred),n_samples))


        time_unnormed = x_pred*cosmo.age(zval).value*1e9

        for i in tqdm(range(n_samples)):
            temp = gp.sample_conditional(yax,x_pred)

            mass_unnormed = temp*np.power(10,mass)
            sfr_unnormed = np.diff(mass_unnormed)/np.diff(time_unnormed)
            samples[1:,i] = sfr_unnormed

        samples[samples<0] = 0

        return samples


    #print("Initial ln-likelihood: {0:.2f}".format(gp.log_likelihood(yax.ravel())))

    result = minimize(neg_ln_like, gp.get_parameter_vector(), jac=grad_neg_ln_like, args=(gp,yax))
    #print(result)

    gp.set_parameter_vector(result.x)
    #print("\nFinal ln-likelihood: {0:.2f}".format(gp.log_likelihood(yax.ravel())))

    y_pred, pred_var = gp.predict(yax.ravel(), x_pred, return_var=True)


    #------------------------------------

    mass_unnormed = y_pred*np.power(10,mass)
    time_unnormed = x_pred*cosmo.age(zval).value*1e9
    #print(np.amax(time_unnormed))

    sfr_unnormed = np.diff(mass_unnormed)/np.diff(time_unnormed)
    gen_sfh = np.zeros((mass_unnormed.shape))
    gen_sfh[1:] = sfr_unnormed

    mask = (gen_sfh<0)
    gen_sfh[mask] = 0



    #y_mean, y_cov = gp.predict(X_[:,np.newaxis], return_cov=True)
    if vb == True:
        plt.figure(figsize=(15,15))
        #plt.plot(input_params_full,input_mass,'k.',markersize=15)
        plt.errorbar(xax,yax,yerr=yerr, markersize=15, marker='o',lw=0,elinewidth=2)
        plt.plot(x_pred,y_pred,'g--')
        plt.fill_between(x_pred,y_pred.ravel()- np.sqrt(pred_var),y_pred.ravel() + np.sqrt(pred_var),alpha=0.1,color='g')
        #plt.axis([0,1,0,1])
        plt.xlabel('normalized time')
        plt.ylabel('normalized mass')
        plt.show()

        plt.plot(np.amax(time_unnormed)/1e9-time_unnormed/1e9,gen_sfh,'k-',lw=3)
        plt.xlabel('t [lookback time; Gyr]',fontsize=14)
        plt.ylabel('SFR(t) [solar masses/yr]',fontsize=14)
        plt.show()

    if std == False:
        return gen_sfh, time_unnormed
    else:
        mass_sigmaup_unnormed = (y_pred.ravel() + np.sqrt(pred_var))*np.power(10,mass)
        mass_sigmadn_unnormed = (y_pred.ravel() - np.sqrt(pred_var))*np.power(10,mass)
        sfr_sigmaup_unnormed = np.diff(mass_sigmaup_unnormed)/np.diff(time_unnormed)
        sfr_sigmadn_unnormed = np.diff(mass_sigmadn_unnormed)/np.diff(time_unnormed)
        gen_sfh_up = np.zeros((mass_unnormed.shape))
        gen_sfh_dn = np.zeros((mass_unnormed.shape))
        gen_sfh_up[1:] = sfr_sigmaup_unnormed
        gen_sfh_dn[1:] = sfr_sigmadn_unnormed

        maskup = (gen_sfh_up < 0)
        maskdn = (gen_sfh_dn < 0)

        gen_sfh_up[maskup] = 0
        gen_sfh_dn[maskdn] = 0
        return gen_sfh, gen_sfh_up, gen_sfh_dn, time_unnormed
    
# older code
    
def gp_sfh_george_v2(sfh_tuple,zval=5.606,vb = False,std=False, sample_posterior = False, n_samples = 10):
    """
    Create a GP-approximation of an SFH based on the george GP package by DFM:
    mass = stellar mass
    sfr = current sfr, averaged over some timescale the SED is sensitive to, 10-100Myr
    Nparam = number of parameters. 1=t50, 2=t33,t67, 3=t25,t50,75 etc
    param_arr = the numpy array of actual parameters

    """

    mass = sfh_tuple[0]
    #sfr = 10**sfh_tuple[1]
    sfr = sfh_tuple[1]
    Nparam = int(sfh_tuple[2])
    param_arr = sfh_tuple[3:]

    # this is the times at which the galaxy formed x percentiles of its observed stellar mass...
    input_Nparams = Nparam
    input_params = param_arr

    # these are time quantities (tx)
    input_params_full = np.zeros((input_Nparams+5,))
    input_params_full[0] = 0
    input_params_full[1] = 0.01
    input_params_full[2:-3] = input_params
    input_params_full[-3] = 0.998
    input_params_full[-2] = 0.999
    input_params_full[-1] = 1.0

    # these are galaxy mass quantities M(t)
    temp_mass = np.linspace(0,1,input_Nparams+2)
    input_mass = np.zeros((input_params_full.shape))

    input_mass[1] = 0.0
    input_mass[2] = 1e-1
    input_mass[2:-3] = temp_mass[1:-1]
    input_mass[-3] = 1.0 - np.power(10,sfr)*(1.0-0.998)*(cosmo.age(zval).value*1e9)/np.power(10,mass)
    input_mass[-2] = 1.0 - np.power(10,sfr)*(1.0-0.999)*(cosmo.age(zval).value*1e9)/np.power(10,mass)
    input_mass[-1] = 1.0
    # the last two statements help to fix the SFR at t_obs

    xax = input_params_full
    yax = (input_mass)
    yerr = np.zeros_like(yax)
    #yerr[1:-3] = yax[1:-3]*0.05
    #yerr[2:-3] = (1-yax[2:-3])*0.05
    #yerr[2:-3] = 0.05
    yerr[2:-3] = 0.001/np.sqrt(Nparam)
    if Nparam > 20:
        yerr[2:-3] = 0.1/np.sqrt(Nparam)

    #------------------------------------

    #kernel = DotProduct(10.0, (1e-2,1e2)) *RationalQuadratic(0.1)
    #gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=9)
    #gp.fit(xax,yax)

    #x = np.linspace(0,1,1000)
    #y_pred, sigma = gp.predict(x[:,np.newaxis], return_std=True)
    #y_pred = y_pred.ravel() + x

    #------------------------------------


    #kernel = np.var(yax) * kernels.ExpSquaredKernel(np.median(yax)+np.std(yax))
    #k2 = np.var(yax) * kernels.LinearKernel(np.median(yax),order=1)
    kernel = np.var(yax) * kernels.Matern32Kernel(np.median(yax)) #+ k2
    gp = george.GP(kernel)

    #print(xax.shape, yerr.shape)

    gp.compute(xax.ravel(), yerr.ravel())

    x_pred = np.linspace(np.amin(xax), np.amax(xax), 1000)
    pred, pred_var = gp.predict(yax.ravel(), x_pred, return_var=True)
    y_pred = pred

    if sample_posterior == True:

        #print('this is happening!')
        samples = np.zeros((len(x_pred),n_samples))


        time_unnormed = x_pred*cosmo.age(zval).value*1e9

        for i in tqdm(range(n_samples)):
            temp = gp.sample_conditional(yax,x_pred)

            mass_unnormed = temp*np.power(10,mass)
            sfr_unnormed = np.diff(mass_unnormed)/np.diff(time_unnormed)
            samples[1:,i] = sfr_unnormed

        samples[samples<0] = 0

        return samples


    #print("Initial ln-likelihood: {0:.2f}".format(gp.log_likelihood(yax.ravel())))

    result = minimize(neg_ln_like, gp.get_parameter_vector(), jac=grad_neg_ln_like, args=(gp,yax))
    #print(result)

    gp.set_parameter_vector(result.x)
    #print("\nFinal ln-likelihood: {0:.2f}".format(gp.log_likelihood(yax.ravel())))

    y_pred, pred_var = gp.predict(yax.ravel(), x_pred, return_var=True)


    #------------------------------------

    mass_unnormed = y_pred*np.power(10,mass)
    time_unnormed = x_pred*cosmo.age(zval).value*1e9
    #print(np.amax(time_unnormed))

    sfr_unnormed = np.diff(mass_unnormed)/np.diff(time_unnormed)
    gen_sfh = np.zeros((mass_unnormed.shape))
    gen_sfh[1:] = sfr_unnormed

    mask = (gen_sfh<0)
    gen_sfh[mask] = 0



    #y_mean, y_cov = gp.predict(X_[:,np.newaxis], return_cov=True)
    if vb == True:
        plt.figure(figsize=(15,15))
        #plt.plot(input_params_full,input_mass,'k.',markersize=15)
        plt.errorbar(xax,yax,yerr=yerr, markersize=15, marker='o',lw=0,elinewidth=2)
        plt.plot(x_pred,y_pred,'g--')
        plt.fill_between(x_pred,y_pred.ravel()- np.sqrt(pred_var),y_pred.ravel() + np.sqrt(pred_var),alpha=0.1,color='g')
        #plt.axis([0,1,0,1])
        plt.xlabel('normalized time')
        plt.ylabel('normalized mass')
        plt.show()

        plt.plot(np.amax(time_unnormed)/1e9-time_unnormed/1e9,gen_sfh,'k-',lw=3)
        plt.xlabel('t [lookback time; Gyr]',fontsize=14)
        plt.ylabel('SFR(t) [solar masses/yr]',fontsize=14)
        plt.show()

    if std == False:
        return gen_sfh, time_unnormed
    else:
        mass_sigmaup_unnormed = (y_pred.ravel() + np.sqrt(pred_var))*np.power(10,mass)
        mass_sigmadn_unnormed = (y_pred.ravel() - np.sqrt(pred_var))*np.power(10,mass)
        sfr_sigmaup_unnormed = np.diff(mass_sigmaup_unnormed)/np.diff(time_unnormed)
        sfr_sigmadn_unnormed = np.diff(mass_sigmadn_unnormed)/np.diff(time_unnormed)
        gen_sfh_up = np.zeros((mass_unnormed.shape))
        gen_sfh_dn = np.zeros((mass_unnormed.shape))
        gen_sfh_up[1:] = sfr_sigmaup_unnormed
        gen_sfh_dn[1:] = sfr_sigmadn_unnormed

        maskup = (gen_sfh_up < 0)
        maskdn = (gen_sfh_dn < 0)

        gen_sfh_up[maskup] = 0
        gen_sfh_dn[maskdn] = 0
        return gen_sfh, gen_sfh_up, gen_sfh_dn, time_unnormed
    
def correct_for_mass_loss(sfh, time, mass_loss_curve_time, mass_loss_curve):

    correction_factors = np.interp(time, mass_loss_curve_time, mass_loss_curve)
    return sfh * correction_factors

def tuple_to_sfh_old_dec11(sfh_tuple, zval, interpolator = 'gp_george', decouple_sfr = False, decouple_sfr_time = 10, vb = False,cosmo = cosmo):
    # generate an SFH from an input tuple (Mass, SFR, {tx}) at a specified redshift


    Nparam = int(sfh_tuple[2])
    mass_quantiles = np.linspace(0,1,Nparam+2)
    time_quantiles = np.zeros_like(mass_quantiles)
    time_quantiles[-1] = 1
    time_quantiles[1:-1] = sfh_tuple[3:]

    # now add SFR constraints

    # SFR smoothly increasing from 0 at the big bang
    mass_quantiles = np.insert(mass_quantiles,1,[0.00])
    time_quantiles = np.insert(time_quantiles,1,[0.01])

    # SFR constrained to SFR_inst at the time of observation
    SFH_constraint_percentiles = np.array([0.96,0.97,0.98,0.99])
    #SFH_constraint_percentiles = np.array([0.97,0.98,0.99])
    for const_vals in SFH_constraint_percentiles:

        delta_mstar = 10**(sfh_tuple[0]) *(1-const_vals)
        delta_t = 1 - delta_mstar/(10**sfh_tuple[1])/(cosmo.age(zval).value*1e9)

        if (delta_t > time_quantiles[-2]) & (delta_t > 0.9):
            mass_quantiles = np.insert(mass_quantiles, -1, [const_vals], )
            time_quantiles = np.insert(time_quantiles, -1, [delta_t],)
        else:
            delta_m = 1 - ((cosmo.age(zval).value*1e9)*(1-const_vals)*(10**sfh_tuple[1]))/(10**sfh_tuple[0])
            time_quantiles = np.insert(time_quantiles, -1, [const_vals])
            mass_quantiles=  np.insert(mass_quantiles, -1, [delta_m])

    if interpolator == 'gp_george':
        time_arr_interp, mass_arr_interp = gp_interpolator(time_quantiles, mass_quantiles, Nparam = int(Nparam), decouple_sfr = decouple_sfr)
    elif interpolator == 'gp_sklearn':
        time_arr_interp, mass_arr_interp = gp_sklearn_interpolator(time_quantiles, mass_quantiles)
    elif interpolator == 'linear':
        time_arr_interp, mass_arr_interp = linear_interpolator(time_quantiles, mass_quantiles)
    elif interpolator == 'pchip':
        time_arr_interp, mass_arr_interp = Pchip_interpolator(time_quantiles, mass_quantiles)
    else:
        raise Exception('specified interpolator does not exist: {}. \n use one of the following: gp_george, gp_sklearn, linear, and pchip '.format(interpolator))

    sfh_scale = 10**(sfh_tuple[0])/(cosmo.age(zval).value*1e9/1000)
    sfh = np.diff(mass_arr_interp)*sfh_scale
    sfh[sfh<0] = 0
    sfh = np.insert(sfh,0,[0])
    if decouple_sfr == True:
        sfr_decouple_time_index = np.argmin(np.abs(time_arr_interp*cosmo.age(zval).value - decouple_sfr_time/1e3))
        sfh[-sfr_decouple_time_index:] = 10**sfh_tuple[1]

    timeax = time_arr_interp * cosmo.age(zval).value

    if vb == True:
        print('time and mass quantiles:')
        print(time_quantiles, mass_quantiles)
        plt.plot(time_quantiles, mass_quantiles,'--o')
        plt.plot(time_arr_interp, mass_arr_interp)
        plt.axis([0,1,0,1])
        #plt.axis([0.9,1.05,0.9,1.05])
        plt.show()

        print('instantaneous SFR: %.1f' %sfh[-1])
        plt.plot(np.amax(time_arr_interp) - time_arr_interp, sfh)
        #plt.xscale('log')
        plt.show()

    return sfh, timeax



import numpy as np
from tqdm import tqdm
import scipy.io as sio

from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.stats import gaussian_kde

# cosmology assumption
from astropy.cosmology import FlatLambdaCDM
cosmo = FlatLambdaCDM(H0=70, Om0=0.3)

from .priors import *
from .pre_grid import *
from .gp_sfh import *
from .plotter import *

class SedFit(object):
    """
    Class to incorporate SED likelihood evaluation and resulting posteriors

    Attributes
    ----------
    mass : log stellar mass
        upper and lower limits of distribution
    sfr : log instantaneous SFR
        upper and lower limits of distribution
    ssfr : log specific SFR = SFR/M*
        negative lognormal distribution parameters
    tx : lookback times at which a galaxy formed certain fractions of its mass
        concentration (tx_alpha) and number of parameters (Nparam) of distribution
    Z  : log metallicity/solar
        upper and lower limits of distribution
    Av : calzetti dust attenuation
        upper and lower limits of distribution
    z  : redshift
        upper and lower limits of distribution

    Methods
    -------
    method: what it does
    """

    def __init__(self, sed, sed_err, atlas, fit_mask = [], zbest = None, deltaz = None):

        self.sed = sed
        self.sed_err = sed_err
        self.atlas = atlas
        self.fit_mask = fit_mask
        self.zbest = zbest
        self.deltaz = deltaz
        self.dynamic_norm = True

    def evaluate_likelihood(self):

        chi2_array, norm_fac = evaluate_sed_likelihood(self.sed, self.sed_err, self.atlas, self.fit_mask, self.zbest, self.deltaz, self.dynamic_norm)
        self.chi2_array = chi2_array
        self.norm_fac = norm_fac
        self.likelihood = np.exp(-(chi2_array)/2)

        return

    def evaluate_posterior_percentiles(self, bw_dex = 0.001, percentile_values = [50.,16.,84.], vb = False):
        """
        by default, the percentile values are median, lower68, upper68.
        change this to whatever the desired sampling of the posterior is.
        """

        quants = get_quants(self.chi2_array, self.atlas, self.norm_fac, bw_dex = bw_dex, percentile_values = percentile_values, vb = vb)

        self.mstar = quants[0]
        self.sfr = quants[1]
        self.Av = quants[2]
        self.Z = quants[3]
        self.z = quants[4]
        self.sfh_tuple = quants[5]
        self.percentile_values = percentile_values

        return

    def evaluate_MAP_mstar(self, bw_dex = 0.001, smooth = 'kde', lowess_frac = 0.3, bw_method = 'scott', vb = False):

        qty = self.atlas['mstar'] + np.log10(self.norm_fac),
        weights = self.likelihood
        bins = np.arange(4,14,bw_dex)
        self.mstar_MAP = evaluate_MAP(qty, weights, bins, smooth = smooth, lowess_frac=lowess_frac, bw_method=bw_method, vb=vb)
        return self.mstar_MAP

    def evaluate_MAP_sfr(self, bw_dex = 0.001, smooth = 'kde', lowess_frac = 0.3, bw_method = 'scott', vb = False):

        qty = self.atlas['sfr'] + np.log10(self.norm_fac),
        weights = self.likelihood
        bins = np.arange(-6,4,bw_dex),
        self.sfr_MAP = evaluate_MAP(qty, weights, bins, smooth = smooth, lowess_frac=lowess_frac, bw_method=bw_method, vb=vb)
        return self.sfr_MAP

    def plot_posteriors(self,truths = []):

        figure = plot_posteriors(self.chi2_array, self.norm_fac, self.sed, self.atlas, truths = truths)
        return figure

    def plot_posterior_spec(self, filt_centers, priors, ngals = 100, alpha=0.1, fnu=True, yscale='log', speccolor = 'k', sedcolor='b', titlestr = [],figsize=(12,7)):

        set_plot_style()

        lam_all = []
        spec_all = []
        z_all = []

        bestn_gals = np.argsort(self.likelihood)

        for i in range(ngals):

            lam_gen, spec_gen =  makespec_atlas(self.atlas, bestn_gals[-(i+1)], priors, mocksp, cosmo, filter_list = [], filt_dir = [], return_spec = True)

            lam_all.append(lam_gen)
            spec_all.append(spec_gen)
            z_all.append(self.atlas['zval'][bestn_gals[-(i+1)]])

        fig = plt.subplots(1,1,figsize=figsize)


        if fnu == True:
            for i in range(ngals):
                plt.plot(lam_all[i]*(1+z_all[i]), spec_all[i]*self.norm_fac, color = speccolor, alpha=alpha)
            plt.errorbar(filt_centers[self.sed>0], self.sed[self.sed>0], yerr=self.sed_err[self.sed>0]*2, color=sedcolor,lw=0, elinewidth=2, marker='o', markersize=12, capsize=5)
            plt.ylabel(r'$F_\nu$ [$\mu$Jy]')

        elif fnu == False:
            for i in range(ngals):
                spec_flam = ujy_to_flam(spec_all[i]*self.norm_fac, lam_all[i]*(1+z_all[i]))
                plt.plot(lam_all[i]*(1+z_all[i]), spec_flam, color = speccolor, alpha=alpha)
            sed_flam = ujy_to_flam(self.sed,filt_centers)
            sed_flam_err_up = ujy_to_flam(self.sed+self.sed_err,filt_centers) - sed_flam
            sed_flam_err_dn = sed_flam - ujy_to_flam(self.sed-self.sed_err,filt_centers)
            # make these F_\lam errors, not F_nu errors
            plt.errorbar(filt_centers[self.sed>0], sed_flam[self.sed>0], yerr=(sed_flam_err_up[self.sed>0], sed_flam_err_dn[self.sed>0]), color=sedcolor,lw=0, elinewidth=2, marker='o', markersize=12, capsize=5)
            plt.ylabel(r'$F_\lambda$')

        plt.xlabel(r'$\lambda$ [$\AA$]')
        plt.xlim(np.amin(filt_centers)*0.81, np.amax(filt_centers)*1.2)
        plt.ylim(np.amin(self.sed[self.sed>0])*0.8,np.amax(self.sed[self.sed>0]+self.sed_err[self.sed>0])*1.5)
        plt.xscale('log');plt.yscale(yscale);
        #plot_lines(filt_centers, gal_z)
        #plt.title(titlestr,fontsize=18)

        return fig

    def evaluate_posterior_SFH(self, zval,ngals=100):

        bestn_gals = np.argsort(self.likelihood)
        common_time = np.linspace(0,cosmo.age(zval).value,100)
        if priors.dynamic_decouple == True:
            priors.decouple_sfr_time = 100*cosmo.age(zval).value/cosmo.age(0.1).value

        all_sfhs = []
        all_weights = []
        for i in (range(ngals)):
            sfh, timeax =  tuple_to_sfh(self.atlas['sfh_tuple'][bestn_gals[-(i+1)],0:],self.atlas['zval'][bestn_gals[-(i+1)]])
            sfh = sfh*self.norm_fac
            sfh_interp = np.interp(common_time, timeax, sfh)

            all_sfhs.append(sfh_interp)
            all_weights.append(self.likelihood[bestn_gals[-(i+1)]])

        all_sfhs = np.array(all_sfhs)
        all_weights = np.array(all_weights)

        sfh_50 = np.zeros_like(common_time)
        sfh_16 = np.zeros_like(common_time)
        sfh_84 = np.zeros_like(common_time)
        for ti in range(len(common_time)):
            qty = np.log10(all_sfhs[0:,ti])
            qtymask = (qty > -np.inf) & (~np.isnan(qty))
            if np.sum(qtymask) > 0:
                smallwts = all_weights.copy()[qtymask.ravel()]
                qty = qty[qtymask]
                if len(qty>0):
                    sfh_50[ti], sfh_16[ti], sfh_84[ti] = 10**calc_percentiles(qty, smallwts, bins=50, percentile_values=[50., 16., 84.])
#             else:
#                 print(common_time[ti])

        return sfh_50, sfh_16, sfh_84, common_time


    def plot_posterior_SFH(self, zval,ngals=100,alpha=0.1, speccolor = 'k', sedcolor='b',figsize=(12,7)):


        set_plot_style()
        bestn_gals = np.argsort(self.likelihood)

        fig = plt.subplots(1,1,figsize=figsize)

        common_time = np.linspace(0,cosmo.age(zval).value,100)
        if priors.dynamic_decouple == True:
            priors.decouple_sfr_time = 100*cosmo.age(zval).value/cosmo.age(0.1).value

        all_sfhs = []
        all_weights = []
        for i in (range(ngals)):
            sfh, timeax =  tuple_to_sfh(self.atlas['sfh_tuple'][bestn_gals[-(i+1)],0:],self.atlas['zval'][bestn_gals[-(i+1)]])
            sfh = sfh*self.norm_fac
            sfh_interp = np.interp(common_time, timeax, sfh)
            all_sfhs.append(sfh_interp)
            all_weights.append(self.likelihood[bestn_gals[-(i+1)]])
            alphawt = 1.0*alpha*self.likelihood[bestn_gals[-(i+1)]]/self.likelihood[bestn_gals[-1]]
            #print(alphawt)
            plt.plot(np.amax(common_time)-common_time, sfh_interp, color = sedcolor, alpha=alphawt)
            #plt.plot(np.amax(timeax) - timeax, sfh*norm_fac, color = speccolor, alpha=alpha)
        all_sfhs = np.array(all_sfhs)
        all_weights = np.array(all_weights)

        sfh_50 = np.zeros_like(common_time)
        sfh_16 = np.zeros_like(common_time)
        sfh_84 = np.zeros_like(common_time)
        for ti in range(len(common_time)):
            qty = np.log10(all_sfhs[0:,ti])
            qtymask = (qty > -np.inf) & (~np.isnan(qty))
            if np.sum(qtymask) > 0:
                smallwts = all_weights.copy()[qtymask.ravel()]
                qty = qty[qtymask]
                if len(qty>0):
                    sfh_50[ti], sfh_16[ti], sfh_84[ti] = 10**calc_percentiles(qty, smallwts, bins=50, percentile_values=[50., 16., 84.])
#             else:
#                 print(common_time[ti])
        plt.plot(np.amax(common_time)-common_time, sfh_50,lw=3,color=speccolor)
        plt.fill_between(np.amax(common_time)-common_time.ravel(), sfh_16.ravel(), sfh_84.ravel(),alpha=0.3,color=speccolor)

        plt.xlabel(r'lookback time [$Gyr$]')
        plt.ylabel(r'$SFR(t)$ [M$_\odot /$yr]')
        try:
            plt.ylim(0,np.amax(sfh_84)*1.2)
        except:
            print('couldnt set axis limits')

        return fig





#-------------------------------------------------------------


def normerr(nf, pg_seds, sed, sed_err, fit_mask):
    c2v = np.amin(np.mean((pg_seds[fit_mask,0:] - sed.reshape(-1,1)/nf)**2 / (sed_err.reshape(-1,1)/nf)**2, 0))
    return c2v

def evaluate_sed_likelihood(sed, sed_err, atlas, fit_mask = [], zbest = None, deltaz = None, dynamic_norm = True):

    """
    Evaluate the likeihood of model SEDs in an atlas given
    an observed SED with uncertainties.
    """

    # preprocessing:
    if len(fit_mask) == len(sed):
        fit_mask = fit_mask & (sed > 0)
    else:
        fit_mask = (sed > 0)

    if len(sed) != len(sed_err):
        raise ValueError('SED uncertainty array does not match SED')

    sed = sed[fit_mask]
    sed_err = sed_err[fit_mask]
    pg_seds = atlas['sed'].copy().T

    if zbest is not None:
        pg_z = atlas['zval'].ravel()
        redshift_mask = (np.abs(pg_z - zbest) < deltaz)
        chi2 = np.zeros((len(pg_z),))
        chi2[~redshift_mask] = 1e10

    if np.sum(redshift_mask) == 0:
        print('atlas does not contain any models in redshift range')

    if dynamic_norm == True:
        nfmin = minimize(normerr, np.nanmedian(sed), args = (pg_seds[0:,redshift_mask], sed, sed_err, fit_mask))
        norm_fac = nfmin['x'][0]
    elif dynamic_norm == False:
        norm_fac = np.nanmedian(sed)
    else:
        norm_fac = 1.0
        print('undefined norm method. using norm_fac = 1')

    sed_normed = sed.reshape(-1,1)/norm_fac
    sed_err_normed = sed_err.reshape(-1,1)/norm_fac

    chi2[redshift_mask] = np.mean((pg_seds[fit_mask,0:][0:,redshift_mask] - sed_normed)**2 / (sed_err_normed)**2, 0)

    # if zbest is not None:
        # pg_z = atlas['zval'].ravel()
        # #redshift_mask = (np.abs(pg_z - zbest) > 0.1*zbest)
        # redshift_mask = (np.abs(pg_z - zbest) > deltaz)
        # chi2[~redshift_mask] = np.amax(chi2)+1e3

    return chi2, norm_fac

def get_quants_key(key, bins, chi2_array, cat, norm_fac, percentile_values = [50.,16.,84.], return_uncert = True, vb = False):
    """
    Get posterior percentiles for an input key
    """
    relprob = np.exp(-(chi2_array)/2)
    key_vals = calc_percentiles(cat[key], weights = relprob, bins = bins, percentile_values = percentile_values, vb = vb)

    return key_vals


def get_quants(chi2_array, cat, norm_fac, bw_dex = 0.001, percentile_values = [50.,16.,84.], vb = False):

    """
    remember to check bin limits and widths before using quantities if you're fitting a new sample
    """

    #relprob = np.exp(-(chi2_array*np.sum(obs_sed>0))/2)
    relprob = np.exp(-(chi2_array)/2)
    if vb == True:
        plt.hist(relprob,100)
        plt.yscale('log')
        plt.show()

    # ---------------- stellar mass and SFR -----------------------------------

    mstar_vals = calc_percentiles(cat['mstar'] + np.log10(norm_fac),
                                 weights = relprob,
                                 bins = np.arange(4,14,bw_dex),
                                 percentile_values = percentile_values, vb=vb)

    sfr_vals = calc_percentiles(cat['sfr'] + np.log10(norm_fac),
                                 weights = relprob,
                                 bins = np.arange(-6,4,bw_dex),
                                 percentile_values = percentile_values,vb=vb)

     # ---------------- SFH -----------------------------------

    sfh_tuple_vals = np.zeros((3, cat['sfh_tuple_rec'].shape[1]))

    for i in range(cat['sfh_tuple_rec'].shape[1]):

        if i == 0:
            sfh_tuple_vals[0:,i] = calc_percentiles(cat['sfh_tuple_rec'][0:,0] + np.log10(norm_fac),
                                 weights = relprob, bins = np.arange(4,14,bw_dex),
                                 percentile_values = percentile_values, vb=vb)
        elif i == 1:
            sfh_tuple_vals[0:,i] = calc_percentiles(cat['sfh_tuple_rec'][0:,1] + np.log10(norm_fac),
                                 weights = relprob, bins = np.arange(-6,4,bw_dex),
                                 percentile_values = percentile_values, vb=vb)
        elif i == 2:
            sfh_tuple_vals[0:,i] = sfh_tuple_vals[0:,i] + np.nanmean(cat['sfh_tuple_rec'][0:,2])
        else:
            sfh_tuple_vals[0:,i] = calc_percentiles(cat['sfh_tuple_rec'][0:,i],
                                 weights = relprob, bins = np.arange(0,1,bw_dex),
                                 percentile_values = percentile_values, vb=vb)

    # ------------------------ dust, metallicity, redshift ----------------------

    Av_vals = calc_percentiles(cat['dust'].ravel(),
                                 weights = relprob,
                                 bins = np.arange(0,np.amax(cat['dust']),bw_dex),
                                 percentile_values = percentile_values,vb=vb)

    Z_vals = calc_percentiles(cat['met'].ravel(),
                                 weights = relprob,
                                 bins = np.arange(-1.5, 0.5,bw_dex),
                                 percentile_values = percentile_values,vb=vb)

    z_vals = calc_percentiles(cat['zval'].ravel(),
                                 weights = relprob,
                                 bins = np.arange(np.amin(cat['zval']), np.amax(cat['zval']),bw_dex),
                                 percentile_values = percentile_values,vb=vb)

    return [mstar_vals, sfr_vals, Av_vals, Z_vals, z_vals, sfh_tuple_vals]

def get_flat_posterior(qty, weights, bins):

    post, xaxis = np.histogram(qty, weights=weights, bins=bins)
    prior, xaxis = np.histogram(qty, bins=bins)

#     post_weighted = post/prior
    post_weighted = post
    post_weighted[np.isnan(post_weighted)] = 0

    return post_weighted, xaxis

def calc_percentiles(qty, weights, bins, percentile_values, vb = False):

    qty_percentile_values = np.zeros((len(percentile_values),))

    post_weighted, xaxis = get_flat_posterior(qty, weights, bins)
    bw = np.nanmean(np.diff(xaxis))

    normed_cdf = np.cumsum(post_weighted)/np.amax(np.cumsum(post_weighted))

    for i in range(len(percentile_values)):
        qty_percentile_values[i] = xaxis[0:-1][np.argmin(np.abs(normed_cdf - percentile_values[i]/100))] + bw/2
        if (qty_percentile_values[i] == xaxis[0]+bw/2) or (qty_percentile_values[i] == xaxis[-1]+bw/2):
            qty_percentile_values[i] = np.nan

    if vb == True:

        qty_50 = xaxis[0:-1][np.argmin(np.abs(normed_cdf - 0.5))] + bw/2
        qty_16 = xaxis[0:-1][np.argmin(np.abs(normed_cdf - 0.16))] + bw/2
        qty_84 = xaxis[0:-1][np.argmin(np.abs(normed_cdf - 0.84))] + bw/2

        plt.plot(xaxis[0:-1]+bw/2, post_weighted)
        bf_value = qty[np.argmax(weights)]
        tempy = plt.ylim()
        plt.plot([bf_value, bf_value],[0,tempy[1]],'k-',label='best-fit')
        plt.plot([qty_50, qty_50],[0,tempy[1]],'-',label='50th percentile')
        plt.plot([qty_16, qty_16],[0,tempy[1]],'-',label='16th percentile')
        plt.plot([qty_84, qty_84],[0,tempy[1]],'-',label='84th percentile')
        print(bf_value, qty_50, np.argmax(weights), np.amax(weights))
        plt.legend(edgecolor='w',fontsize=14)
        plt.show()

    return qty_percentile_values


def evaluate_MAP(qty, weights, bins, smooth = 'kde', lowess_frac = 0.3, bw_method = 'scott', vb = False):

    post, xaxis = np.histogram(qty, weights=weights, bins=bins)
    xaxis_centers = xaxis[0:-1] + np.mean(np.diff(xaxis))

    if smooth == 'lowess':
        a = lowess(post, xaxis_centers,frac=lowess_frac)
        MAP = a[np.argmax(a[0:,1]),0]
    elif smooth == 'kde':
        a = gaussian_kde(qty,bw_method=bw_method, weights=weights)
        MAP = xaxis[np.argmax(a.evaluate(xaxis))]
    else:
        MAP = xaxis[np.argmax(post)+1]

    if vb == True:
        areapost = np.trapz(x=xaxis_centers, y=post)
        plt.plot(xaxis_centers, post/areapost)
        if smooth == 'lowess':
            plt.plot(a[0:,0],a[0:,1]/areapost)
        elif smooth == 'kde':
            plt.plot(xaxis, a.pdf(xaxis))
        plt.plot([MAP,MAP],plt.ylim())
        plt.show()

    return MAP


def get_lines(centers, zval):

    # line list from Table 4 of https://iopscience.iop.org/article/10.3847/1538-4357/aa6c66 (Byler+17)

    lam_min = np.amin(centers/(1+zval))*0.81
    lam_max = np.amax(centers/(1+zval))*1.1

    #line_lam = [1215.6701, 6564.6, 4862.71, 4341.692, 4102.892, 18756.4]
    #line_name = [r'Ly$\alpha$',r'H$\alpha$',r'H$\beta$',r'H$\gamma$',r'H$\delta$',r'Pa$\alpha$']

    #line_lam = [1215.6701, 6564.6, 4862.71, 4102.892, 18756.4,4472.735,1640.42, 9852.9, 8729.53,4622.864,5201.705,6585.27,5756.19, 5008.24, 3727.1,3869.86,40522.79]
    #line_name = [r'Ly$\alpha$',r'H$\alpha$',r'H$\beta$',r'H$\delta$',r'Pa$\alpha$','HeI','HeII','[CI]','[CI]', '[CI]','[NI]','[NII]','[NII]','[OIII]','[OII]','[NeIII]',r'Br$\alpha$']
    #line_offset = [0.0,0.0,0.0,0.0,0.0,0.007,0.007,0.014,0.014,0.014,0.007,0.007,0.007,0.021,0.021,0.028,0.0]

    line_lam = [  1215.67,   1906.68,   3727.1 ,  3869.86 ,  4102.89,
   4862.71 ,  4960.3 ,  6564.6 ,   9071.1,    9533.2,
  10832.06,  10941.17 , 12821.58,  18756.4,   40522.79, 105105.,
 155551.,   187130.,   334800.,   518145.,   883564.  ]
    line_name = [r'Ly$\alpha$','[CIII]','[OII]$_{x2}$','[NeIII]',r'H$\delta$',
                r'H$\beta$','[OIII]$_{x2}$',r'H$\alpha$','[SIII]','[SIII]',
                'He I',r'Pa$\gamma$',r'Pa$\beta$', r'Pa$\alpha$',r'Br$\alpha$','[S IV]',
                '[Ne III]','[S III]','[S III]','[O III]','[O III]']
    line_offset = [0.0,0.007,0.007,0.014,0.000,
                   0.0,0.007,0.0,0.007,0.014,
                  0.007,0.0,0.0,0.0,0.0,0.0,0.0,
                  0.0,0.007,0.014,0.0,0.007]

    good_line_lams, good_line_names, good_line_offsets = [], [], []
    for i in range(len(line_lam)):
        if (line_lam[i] > lam_min) & (line_lam[i] < lam_max):
            good_line_lams.append(line_lam[i])
            good_line_names.append(line_name[i])
            good_line_offsets.append(line_offset[i])
    return good_line_lams, good_line_names, good_line_offsets


def plot_lines(filt_centers, zval,color = 'forestgreen',alpha=0.3,fontsize=14):

    lls, lns, los = get_lines(filt_centers, zval)
    tempy = plt.ylim()
    for i in range(len(lls)):
        plt.plot([lls[i]*(1+zval),lls[i]*(1+zval)],tempy,':',color=color,alpha=alpha)
        plt.text(lls[i]*(1+zval)*1.03, (0.003+los[i])*(tempy[1] - tempy[0]) + tempy[0],lns[i],fontsize=fontsize,color=color)
    plt.ylim(tempy)
    return

def ujy_to_flam(data,lam):
    flam = ((3e-5)*data)/((lam**2.)*(1e6))
    return flam/1e-19

#---------------------------------------------
#----------- deprecated functions ------------
#---------------------------------------------

def fit_sed_pregrid_old(sed, sed_err, pg_theta, fit_mask = [True], fit_method = 'chi2', norm_method = 'none', return_val = 'params', make_posterior_plots = False, make_sed_plot = False, truths = [np.nan], zbest = None, deltaz = None):
    """
    DEPRECATED
    """
    # preprocessing:
    if len(fit_mask) == len(sed):
        fit_mask = fit_mask & (sed > 0)
    else:
        fit_mask = (sed > 0)

    if len(sed) != len(sed_err):
        raise ValueError('SED error array does not match SED')

    sed = sed[fit_mask]
    sed_err = sed_err[fit_mask]

    pg_sfhs, pg_Z, pg_Av, pg_z, pg_seds = pg_theta

    if fit_method == 'chi2':

        if norm_method == 'none':
            # no normalization
            norm_fac = 1
        elif norm_method == 'max':
            # normalize SEDs to 1
            norm_fac = np.amax(sed)
        elif norm_method == 'median':
            # normalize SEDs to median
            norm_fac = np.median(sed)
        elif norm_method == 'area':
            # normalize SFH to 10^9 Msun - currently not implemented
            norm_fac == 1
        elif norm_method == 'pg_band':
            # normalize to band that pg_seds peak in
            norm_fac = 1
        else:
            raise ValueError('undefined normalization argument')

        sed_normed = sed.reshape(-1,1)/norm_fac
        sed_err_normed = sed_err.reshape(-1,1)/norm_fac

        # fitting the SED
        chi2 = np.mean((pg_seds[fit_mask,0:] - sed_normed)**2 / (sed_err_normed)**2, 0)

        if norm_method == 'pg_band':
            # which bands do the SEDs peak in?
            chi2 = np.ones_like(chi2)*1e3
            pg_sed_peaks = np.argmax(pg_seds,0)
            for tempi in range(len(sed)):
                if fit_mask[i] == True:
                    temp_norm_fac = sed[i]
                    temp_sed_normed = sed.reshape(-1,1)/temp_norm_fac
                    temp_err_normed = sed_err.reshape(-1,1)/temp_norm_fac
                    chi2[pg_sed_peaks == i] = np.mean((pg_seds[fit_mask,pg_sed_peaks == i] - sed_normed)**2 / (sed_err_normed)**2, 0)
            best_chi2 = np.argmin(chi2)
            norm_fac = sed[pg_sed_peaks[best_chi2]]

        if zbest is not None:
            #redshift_mask = (np.abs(pg_z - zbest) > 0.1*zbest)
            redshift_mask = (np.abs(pg_z - zbest) > deltaz)
            chi2[redshift_mask] = np.amax(chi2)+1e3


        if make_sed_plot == True:
            plt.plot(sed_normed)
            plt.plot(sed_normed+sed_err_normed,'k--')
            plt.plot(sed_normed-sed_err_normed,'k--')
            plt.plot(pg_seds[fit_mask,np.argmin(chi2)])
            plt.show()

        if make_posterior_plots == True:
            plt.figure(figsize=(27,4))
            plt.subplot(1,5,1);plt.xlabel('log Stellar Mass')
            plt.hist(pg_sfhs[0,0:]+np.log10(norm_fac),10, weights=np.exp(-chi2/2), normed=True)
            plt.subplot(1,5,2);plt.xlabel('log SFR')
            plt.hist(pg_sfhs[1,0:]+np.log10(norm_fac),10, weights=np.exp(-chi2/2), normed=True)
            plt.subplot(1,5,3);plt.xlabel(r't$_{50}$')
            plt.hist(pg_sfhs[3,0:],10, weights=np.exp(-chi2/2), normed=True)
            plt.subplot(1,5,4);plt.xlabel(r'A$_v$')
            plt.hist(pg_Av,10, weights=np.exp(-chi2/2), normed=True)
            plt.subplot(1,5,5);plt.xlabel(r'log Z/Z$_{\odot}$')
            plt.hist(pg_Z,10, weights=np.exp(-chi2/2), normed=True)
            if len(truths) > 1:
                plt.subplot(1,5,1);tempy = plt.ylim();plt.plot([truths[0],truths[0]],tempy,'k--', lw=3)
                plt.subplot(1,5,2);tempy = plt.ylim();plt.plot([truths[1],truths[1]],tempy,'k--', lw=3)
                plt.subplot(1,5,3);tempy = plt.ylim();plt.plot([truths[2],truths[2]],tempy,'k--', lw=3)
                plt.subplot(1,5,4);tempy = plt.ylim();plt.plot([truths[3],truths[3]],tempy,'k--', lw=3)
                plt.subplot(1,5,5);tempy = plt.ylim();plt.plot([truths[4],truths[4]],tempy,'k--', lw=3)
            plt.show()

        if return_val == 'chi2':
            return chi2
        elif return_val == 'params':
            return calculate_50_16_84_params(chi2, len(sed), np.vstack([pg_sfhs, pg_Av, pg_Z, pg_z]), norm_fac)
        elif return_Val == 'posteriors':
            return calculate_50_16_84_params(chi2, len(sed), np.vstack([pg_sfhs, pg_Av, pg_Z, pg_z]), norm_fac, return_posterior = True)
        else:
            raise ValueError('Undefined return request. Use chi2, params or posteriors.')

    elif fit_method == 'dot':
        dot_prods = np.linalg.multi_dot((pg_seds[fit_mask,0:].T, sed[fit_mask]/np.amax(sed)))/np.linalg.norm(pg_seds,axis=0)
        best_dotprod = np.argmax(dot_prods)
        return 1 - dot_prods/np.amax(dot_prods)

    else:
        print('unknown fit_method - use chi2 or dot.')
        return 0

def get_sfr_mstar(chi2_array, pg_theta, obs_sed, bw_dex = 0.01, return_uncert = True):
    """
    DEPRECATED
    """


    a,b = np.histogram(pg_theta[0][0,0:] + np.log10(np.amax(obs_sed)),
                       weights = np.exp(-chi2_array/2),
                       bins = np.arange(8,12,bw_dex))
    mstar_50 = b[0:-1][np.argmin(np.abs(np.cumsum(a)/np.amax(np.cumsum(a)) - 0.5))] + bw_dex/2
    mstar_16 = b[0:-1][np.argmin(np.abs(np.cumsum(a)/np.amax(np.cumsum(a)) - 0.16))] + bw_dex/2
    mstar_84 = b[0:-1][np.argmin(np.abs(np.cumsum(a)/np.amax(np.cumsum(a)) - 0.84))] + bw_dex/2

    a,b = np.histogram(pg_theta[0][1,0:] + np.log10(np.amax(obs_sed)),
                       weights = np.exp(-chi2_array/2),
                       bins = np.arange(-3,3,bw_dex))
    sfr_50 = b[0:-1][np.argmin(np.abs(np.cumsum(a)/np.amax(np.cumsum(a)) - 0.5))] + bw_dex/2
    sfr_16 = b[0:-1][np.argmin(np.abs(np.cumsum(a)/np.amax(np.cumsum(a)) - 0.16))] + bw_dex/2
    sfr_84 = b[0:-1][np.argmin(np.abs(np.cumsum(a)/np.amax(np.cumsum(a)) - 0.84))] + bw_dex/2

    if return_uncert == False:
        return mstar_50, sfr_50
    else:
        return [mstar_50, mstar_16, mstar_84], [sfr_50, sfr_16, sfr_84]

def calculate_50_16_84_params(chi2, nbands, params, norm_fac, nbins = 30, return_posterior = False):
    """
    DEPRECATED
    """

    relprob = -0.5*chi2
    relprob = relprob - np.amax(relprob)
    relprob = np.exp(relprob/nbands)

    pg_params = params.copy()

    param_arr = pg_params[0,0:] + np.log10(norm_fac)
    param_arr[param_arr < 0] = 0

    [n,bins,] = np.histogram(param_arr,weights=relprob,density=True,bins=nbins)
    n_c = np.cumsum(n)
    n_c = n_c / np.amax(n_c)
    bin_centers = bins[0:-1] + (bins[1]-bins[0])/2

    mass_median = bin_centers[np.argmin(np.abs(n_c - 0.5))]
    mass_16 = bin_centers[np.argmin(np.abs(n_c - 0.16))]
    mass_84 = bin_centers[np.argmin(np.abs(n_c - 0.84))]
    mass_posterior = [n,bins]

    [n,bins,] = np.histogram(pg_params[1,0:] + np.log10(norm_fac),weights=relprob,density=True,bins=nbins)
    n_c = np.cumsum(n)
    n_c = n_c / np.amax(n_c)
    bin_centers = bins[0:-1] + (bins[1]-bins[0])/2

    sfr_median = bin_centers[np.argmin(np.abs(n_c - 0.5))]
    sfr_16 = bin_centers[np.argmin(np.abs(n_c - 0.16))]
    sfr_84 = bin_centers[np.argmin(np.abs(n_c - 0.84))]
    sfr_posterior = [n,bins]

    [n,bins,] = np.histogram(pg_params[-3,0:],weights=relprob,density=True,bins=nbins)
    n_c = np.cumsum(n)
    n_c = n_c / np.amax(n_c)
    bin_centers = bins[0:-1] + (bins[1]-bins[0])/2

    dust_median = bin_centers[np.argmin(np.abs(n_c - 0.5))]
    dust_16 = bin_centers[np.argmin(np.abs(n_c - 0.16))]
    dust_84 = bin_centers[np.argmin(np.abs(n_c - 0.84))]
    dust_posterior = [n,bins]

    [n,bins,] = np.histogram(pg_params[-2,0:],weights=relprob,density=True,bins=nbins)
    n_c = np.cumsum(n)
    n_c = n_c / np.amax(n_c)
    bin_centers = bins[0:-1] + (bins[1]-bins[0])/2

    met_median = bin_centers[np.argmin(np.abs(n_c - 0.5))]
    met_16 = bin_centers[np.argmin(np.abs(n_c - 0.16))]
    met_84 = bin_centers[np.argmin(np.abs(n_c - 0.84))]
    met_posterior = [n,bins]

    times_median = np.zeros(((pg_params.shape[0] - 6),))
    times_16 = np.zeros(((pg_params.shape[0] - 6),))
    times_84 = np.zeros(((pg_params.shape[0] - 6),))
    times_posterior = []
    for i in range(len(times_median)):
        [n,bins,] = np.histogram(pg_params[3+i,0:],weights=relprob,density=True,bins=nbins)
        n_c = np.cumsum(n)
        n_c = n_c / np.amax(n_c)
        bin_centers = bins[0:-1] + (bins[1]-bins[0])/2
        times_median[i] = bin_centers[np.argmin(np.abs(n_c - 0.5))]
        times_16[i] = bin_centers[np.argmin(np.abs(n_c - 0.16))]
        times_84[i] = bin_centers[np.argmin(np.abs(n_c - 0.84))]
        times_posterior.append([n, bins])

    if return_posterior == False:
        return mass_median, mass_16, mass_84, sfr_median, sfr_16, sfr_84, dust_median, dust_16, dust_84, met_median, met_16, met_84, times_median, times_16, times_84
    else:
        return mass_posterior, sfr_posterior, dust_posterior, met_posterior, times_posterior



try:
    from schwimmbad import SerialPool, MultiPool
    from functools import partial
except:
    print('running without parallelization.')
import numpy as np
import os
import time

from astropy.table import Table
import time
import pylab as pl
from IPython import display

from .gp_sfh import *
from .sed_fitter import *
from .pre_grid import *
from .priors import *


def gen_pg_parallel(data_i, atlas_vals):
    """
    Generate the i-th chunk in parallel
    """

    fname, priors, pg_folder, filter_list, filt_dir, N_pregrid = atlas_vals
    fname_full = fname + '_chunk_%.0f' %(data_i)

    generate_atlas(N_pregrid = N_pregrid,
                          priors = priors,
                          fname = fname_full, store=True, path=pg_folder,
                          filter_list = filter_list, filt_dir = filt_dir,
                          rseed = (N_pregrid * data_i + 1))
    return

def generate_atlas_in_parallel_chunking(chunksize, nchunks, fname = 'temp_parallel_atlas', filter_list = 'filter_list_goodss.dat', filt_dir = 'internal', priors = [], pg_folder = 'pregrids/'):
    """
    Generate chunks of an atlas in parallel and combine them into one big atlas
    """

    chunk_path = pg_folder+'atlaschunks/'
    store_path = pg_folder


    atlas_vals = [fname, priors, chunk_path, filter_list, filt_dir, chunksize]

    time_start = time.time()
    data = np.arange(nchunks)

    try:
        with MultiPool() as pool:
            values = list(pool.map(partial(gen_pg_parallel, atlas_vals = atlas_vals), data))
    finally:
        print('Generated pregrid (%.0f chunks, %.0f sedsperchunk)')
        print('time taken: %.2f mins.' %((time.time()-time_start)/60))

    combine_pregrid_chunks(fname_base=fname,
                           N_chunks=nchunks, N_pregrid=chunksize, N_param=priors.Nparam,
                           chunk_path=chunk_path, store_path=store_path)

    return

def combine_pregrid_chunks(fname_base, N_chunks, N_pregrid, N_param, chunk_path='pregrids/atlaschunks/', store_path='pregrids/'):
    """
    A function to combine chunks of a large pregrid generated in parallel
    """

    i=0
    atlas_big = db.load_atlas(fname = fname_base+'_chunk_%.0f' %i, N_pregrid = chunksize, N_param = N_param, path=chunk_path)
    atlas_keys = atlas_big.keys()

    for i in range(1,N_chunks):

        atlas = db.load_atlas(fname = fname_base+'_chunk_%.0f' %i, N_pregrid = chunksize, N_param = N_param, path=chunk_path)

        for key in atlas.keys():

            try:
                if key != 'norm_method':
                    if len(atlas[key].shape) > 1:
                        cmbd_qty = np.vstack((atlas_big[key],atlas[key]))
                    else:
                        cmbd_qty = np.hstack((atlas_big[key],atlas[key]))

                    atlas_big[key] = cmbd_qty
            except:
                print('didnt combine ',key)

    fname = fname_base + '_combined'

    totsize = N_pregrid*N_chunks
    if os.path.exists(store_path):
        print('Path exists. Saved atlas at : '+store_path+fname+'_'+str(N_pregrid * N_chunks)+'_Nparam_'+str(N_param)+'.dbatlas')
    else:
        os.mkdir(store_path)
        print('Created directory and saved atlas at : '+store_path+fname+'_'+str(N_pregrid * N_chunks)+'_Nparam_'+str(N_param)+'.dbatlas')
    try:
        hickle.dump(atlas_big,
                    store_path+fname+'_'+str(N_pregrid * N_chunks)+'_Nparam_'+str(N_param)+'.dbatlas',
                    compression='gzip', compression_opts = 9)
    except:
        print('storing without compression')
        hickle.dump(atlas_big,
                    store_path+fname+'_'+str(N_pregrid * N_chunks)+'_Nparam_'+str(N_param)+'.dbatlas')

    return

# ------------ deprecated -----------------

def make_atlas_parallel(zval, atlas_params):
    """
    Make a single atlas given a redshift value and a list of parameters (including a priors object).
    Atlas Params: [N_pregrid, priors, fname, store, path, filter_list, filt_dir, z_bw]
    """

    # currently only works for photometry, change to include a variable list of atlas_kwargs
    N_pregrid, priors, fname, store, path, filter_list, filt_dir, z_bw = atlas_params

    priors.z_max = zval + z_bw/2
    priors.z_min = zval - z_bw/2

    fname = fname+'_zval_%.0f_' %(zval*10000)

    generate_atlas(N_pregrid = N_pregrid,
                      priors = priors,
                      fname = fname, store=True, path=path,
                      filter_list = filter_list, filt_dir = filt_dir,
                      rseed = int(zval*100000))

    return


def generate_atlas_in_parallel_zgrid(zgrid, atlas_params, dynamic_decouple = True):
    """
    Make a set of atlases given a redshift grid and a list of parameters (including a priors object).
    Atlas Params: [N_pregrid, priors, fname, store, path, filter_list, filt_dir, z_bw]
    """

    time_start = time.time()

    try:
        with MultiPool() as pool:
            values = list(pool.map(partial(make_atlas_parallel, atlas_params = atlas_params), zgrid))
    finally:
        time_end = time.time()
        print('time taken [parallel]: %.2f min.' %((time_end-time_start)/60))


#------------------------------------------------------------------------------------

def fit_gals(gal_id, catvals):


    #if not fit_mask:
    if len(catvals) == 3:
        cat_seds, cat_errs, atlas = catvals
        fit_mask = []

    elif len(catvals) == 4:
        cat_seds, cat_errs, fit_mask, atlas = catvals

    else:
        print('wrong number of arguments supplied to fitter')

    gal_sed = cat_seds[gal_id, 0:].copy()
    gal_err = cat_errs[gal_id, 0:].copy()
    #gal_err = cat_errs[gal_id, 0:].copy() + gal_sed*0.03
    #gal_err = cat_errs[gal_id, 0:].copy() + gal_sed*0.1
    #gal_err = cat_errs[gal_id, 0:].copy() + gal_sed*0.5
    fit_likelihood, fit_norm_fac = evaluate_sed_likelihood(gal_sed,gal_err,atlas,fit_mask=fit_mask,
                                            zbest=None,deltaz=None)

    quants = get_quants(fit_likelihood, atlas, fit_norm_fac)

    return quants, fit_likelihood

#     try:
#         map_mstar = evaluate_MAP(atlas['mstar']+np.log10(fit_norm_fac),
#                                  np.exp(-fit_likelihood/2),
#                                  bins = np.arange(4,14,0.001),
#                                  smooth = 'kde', lowess_frac = 0.3, vb = False)

#         map_sfr = evaluate_MAP(atlas['sfr']+np.log10(fit_norm_fac),
#                                  np.exp(-fit_likelihood/2),
#                                  bins = np.arange(-6,4,0.001),
#                                  smooth = 'kde', lowess_frac = 0.3, vb = False)
#         return quants, fit_likelihood, map_mstar, map_sfr
#     except:
#         print('couldnt calculate MAP for galid: ',gal_id)
#         return quants, fit_likelihood, np.nan, np.nan



def fit_catalog(fit_cat, atlas_path, atlas_fname, output_fname, N_pregrid = 10000, N_param = 3, z_bw = 0.05, f160_cut = 100, fit_mask = [], zgrid = [], sfr_uncert_cutoff = 2.0):

    cat_id, cat_zbest, cat_seds, cat_errs, cat_f160, cat_class_star = fit_cat

    #if not zgrid:
    if isinstance(zgrid, (np.ndarray)) == False:
        zgrid = np.arange(np.amin(cat_zbest),np.amax(cat_zbest),z_bw)

    fit_id = cat_id.copy()
    fit_logM_50 = np.zeros_like(cat_zbest)
    fit_logM_MAP = np.zeros_like(cat_zbest)
    fit_logM_16 = np.zeros_like(cat_zbest)
    fit_logM_84 = np.zeros_like(cat_zbest)
    fit_logSFRinst_50 = np.zeros_like(cat_zbest)
    fit_logSFRinst_MAP = np.zeros_like(cat_zbest)
    fit_logSFRinst_16 = np.zeros_like(cat_zbest)
    fit_logSFRinst_84 = np.zeros_like(cat_zbest)

    fit_logZsol_50 = np.zeros_like(cat_zbest)
    fit_logZsol_16 = np.zeros_like(cat_zbest)
    fit_logZsol_84 = np.zeros_like(cat_zbest)
    fit_Av_50 = np.zeros_like(cat_zbest)
    fit_Av_16 = np.zeros_like(cat_zbest)
    fit_Av_84 = np.zeros_like(cat_zbest)

    fit_zfit_50 = np.zeros_like(cat_zbest)
    fit_zfit_16 = np.zeros_like(cat_zbest)
    fit_zfit_84 = np.zeros_like(cat_zbest)
    fit_logMt_50 = np.zeros_like(cat_zbest)
    fit_logMt_16 = np.zeros_like(cat_zbest)
    fit_logMt_84 = np.zeros_like(cat_zbest)
    fit_logSFR100_50 = np.zeros_like(cat_zbest)
    fit_logSFR100_16 = np.zeros_like(cat_zbest)
    fit_logSFR100_84 = np.zeros_like(cat_zbest)
    fit_nparam = np.zeros_like(cat_zbest)
    fit_t25_50 = np.zeros_like(cat_zbest)
    fit_t25_16 = np.zeros_like(cat_zbest)
    fit_t25_84 = np.zeros_like(cat_zbest)
    fit_t50_50 = np.zeros_like(cat_zbest)
    fit_t50_16 = np.zeros_like(cat_zbest)
    fit_t50_84 = np.zeros_like(cat_zbest)
    fit_t75_50 = np.zeros_like(cat_zbest)
    fit_t75_16 = np.zeros_like(cat_zbest)
    fit_t75_84 = np.zeros_like(cat_zbest)

    fit_nbands = np.zeros_like(cat_zbest)
    fit_f160w = np.zeros_like(cat_zbest)
    fit_stellarity = np.zeros_like(cat_zbest)
    fit_chi2 = np.zeros_like(cat_zbest)
    fit_flags = np.zeros_like(cat_zbest)

    for i in (range(len(zgrid))):

        print('loading atlas at', zgrid[i])

        # for a given redshift slice,
        zval = zgrid[i]

        # select the galaxies to be fit
        z_mask = (cat_zbest < (zval + z_bw/2)) & (cat_zbest > (zval - z_bw/2)) & (cat_f160 < f160_cut)
        fit_ids = np.arange(len(cat_zbest))[z_mask]


#         for gal_id in fit_ids:

#             gal_sed = cat_seds[gal_id, 0:]
#             gal_err = cat_errs[gal_id, 0:]

#             fit_likelihood, fit_norm_fac = evaluate_sed_likelihood(gal_sed,gal_err,atlas,fit_mask=[],
#                                                 zbest=None,deltaz=None)

#             quants = get_quants(fit_likelihood, atlas, fit_norm_fac)

        print('starting parallel fitting for Ngals = ',len(fit_ids),' at redshift ', str(zval))

        try:
#             load the atlas
            fname = atlas_fname+'_zval_%.0f_' %(zgrid[i]*10000)
            atlas = load_atlas(fname, N_pregrid, N_param = N_param, path = atlas_path)
            print('loaded atlas')
            with MultiPool() as pool:
                # note: Parallel doesn't work in Python2.6

#                 if not fit_mask:
                if isinstance(fit_mask, np.ndarray) == False:
                    all_quants = list(pool.map(partial(fit_gals, catvals=(cat_seds, cat_errs, atlas)), fit_ids))
                else:
                    all_quants = list(pool.map(partial(fit_gals, catvals=(cat_seds, cat_errs, fit_mask, atlas)), fit_ids))
            print('finished fitting parallel zbest chunk at z=%.3f' %zval)

            print('starting to put values in arrays')
            for ii, gal_id in enumerate(fit_ids):

                gal_sed = cat_seds[gal_id, 0:]
                gal_err = cat_errs[gal_id, 0:]

                quants = all_quants[ii][0]
                fit_likelihood = all_quants[ii][1]
    #             fit_logM_MAP[gal_id] = all_quants[ii][2]
    #             fit_logSFRinst_MAP[gal_id] = all_quants[ii][3]

                fit_logM_50[gal_id] = quants[0][0]
                fit_logM_16[gal_id] = quants[0][1]
                fit_logM_84[gal_id] = quants[0][2]
                fit_logSFRinst_50[gal_id] = quants[1][0]
                fit_logSFRinst_16[gal_id] = quants[1][1]
                fit_logSFRinst_84[gal_id] = quants[1][2]

                fit_Av_50[gal_id] = quants[2][0]
                fit_Av_16[gal_id] = quants[2][1]
                fit_Av_84[gal_id] = quants[2][2]

                fit_logZsol_50[gal_id] = quants[3][0]
                fit_logZsol_16[gal_id] = quants[3][1]
                fit_logZsol_84[gal_id] = quants[3][2]

                fit_zfit_50[gal_id] = quants[4][0]
                fit_zfit_16[gal_id] = quants[4][1]
                fit_zfit_84[gal_id] = quants[4][2]

                fit_logMt_50[gal_id] = quants[5][0][0]
                fit_logMt_16[gal_id] = quants[5][1][0]
                fit_logMt_84[gal_id] = quants[5][2][0]
                fit_logSFR100_50[gal_id] = quants[5][0][1]
                fit_logSFR100_16[gal_id] = quants[5][1][1]
                fit_logSFR100_84[gal_id] = quants[5][2][1]
                fit_nparam[gal_id] = quants[5][0][2]
                fit_t25_50[gal_id] = quants[5][0][3]
                fit_t25_16[gal_id] = quants[5][1][3]
                fit_t25_84[gal_id] = quants[5][2][3]
                fit_t50_50[gal_id] = quants[5][0][4]
                fit_t50_16[gal_id] = quants[5][1][4]
                fit_t50_84[gal_id] = quants[5][2][4]
                fit_t75_50[gal_id] = quants[5][0][5]
                fit_t75_16[gal_id] = quants[5][1][5]
                fit_t75_84[gal_id] = quants[5][2][5]

                fit_nbands[gal_id] = np.sum(gal_sed>0)
                fit_f160w[gal_id] = cat_f160[gal_id]
                fit_stellarity[gal_id] = cat_class_star[gal_id]
                fit_chi2[gal_id] = np.amin(fit_likelihood)

                # flagging galaxies that either
                # 1. have nan values for mass
                # 2. have SFR uncertainties > sfr_uncert_cutoff
                # 3. are flagged as a star
                # 4. have extremely large chi2
                if np.isnan(quants[0][0]):
                    fit_flags[gal_id] = 1.0
                elif (np.abs(fit_logSFRinst_84[gal_id] - fit_logSFRinst_16[gal_id]) > sfr_uncert_cutoff):
                    fit_flags[gal_id] = 2.0
                elif (cat_class_star[gal_id] > 0.5):
                    fit_flags[gal_id] = 3.0
                elif (fit_chi2[gal_id] > 1000):
                    fit_flags[gal_id] = 4.0
                else:
                    fit_flags[gal_id] = 0.0

        except:
            print('couldn\'t fit with pool at z=',zval)

        print('finishing that')
        pl.clf()
        pl.figure(figsize=(12,6))
        pl.hist(cat_zbest[cat_zbest>0],np.arange(0,6,z_bw),color='black',alpha=0.3)
        #pl.hist(fit_zfit_50[fit_zfit_50>0],np.arange(0,6,z_bw),color='royalblue')
        pl.hist(cat_zbest[fit_zfit_50>0],np.arange(0,6,z_bw),color='royalblue')
        pl.title('fit %.0f/%.0f galaxies' %(np.sum(fit_zfit_50>0), len(cat_zbest)))
        pl.xlabel('redshift');pl.ylabel('# galaxies')

        display.clear_output(wait=True)
        display.display(pl.gcf())

    pl.show()

    #'logSFRinst_MAP':fit_logSFRinst_MAP,
    #'logM_MAP':fit_logM_MAP,

    fit_mdict = {'ID':fit_id,
                 'logM_50':fit_logM_50, 'logM_16':fit_logM_16,'logM_84':fit_logM_84,
                 'logSFRinst_50':fit_logSFRinst_50, 'logSFRinst_16':fit_logSFRinst_16, 'logSFRinst_84':fit_logSFRinst_84,
                 'logZsol_50':fit_logZsol_50, 'logZsol_16':fit_logZsol_16, 'logZsol_84':fit_logZsol_84,
                 'Av_50':fit_Av_50, 'Av_16':fit_Av_16, 'Av_84':fit_Av_84,
                 'zfit_50':fit_zfit_50, 'zfit_16':fit_zfit_16, 'zfit_84':fit_zfit_84,
                 'logMt_50':fit_logMt_50, 'logMt_16':fit_logMt_16, 'logMt_84':fit_logMt_84,
                 'logSFR100_50':fit_logSFR100_50, 'logSFR100_16':fit_logSFR100_16, 'logSFR100_84':fit_logSFR100_84,
                 't25_50':fit_t25_50, 't25_16':fit_t25_16, 't25_84':fit_t25_84,
                 't50_50':fit_t50_50, 't50_16':fit_t50_16, 't50_84':fit_t50_84,
                 't75_50':fit_t75_50, 't75_16':fit_t75_16, 't75_84':fit_t75_84,
                 'nparam':fit_nparam,
                 'nbands':fit_nbands,
                 'F160w':fit_f160w,
                 'stellarity':fit_stellarity,
                 'chi2': fit_chi2,
                 'fit_flags':fit_flags}

    fit_cat = Table(fit_mdict)

    fit_cat.write(output_fname, format='ascii.commented_header')

    return



import numpy as np
import matplotlib.pyplot as plt
import corner

import seaborn as sns

from pylab import *

from .pre_grid import make_filvalkit_simple, load_atlas
from .gp_sfh import *

def set_plot_style():
    sns.set(font_scale=2)
    sns.set_style('ticks')

    rc('axes', linewidth=3)
    rcParams['xtick.major.size'] = 12
    rcParams['ytick.major.size'] = 12
    rcParams['xtick.minor.size'] = 9
    rcParams['ytick.minor.size'] = 9
    rcParams['xtick.major.width'] = 3
    rcParams['ytick.major.width'] = 3

def plot_sfh(timeax, sfh, lookback = False, logx = False, logy = False, fig = None, label=None, **kwargs):
    set_plot_style()

    if fig is None:
        fig = plt.figure(figsize=(12,4))
    if lookback == True:
        plt.plot(np.amax(timeax) - timeax, sfh, label=label, **kwargs)
        plt.xlabel('lookback time [Gyr]');
    else:
        plt.plot(timeax, sfh, label=label, **kwargs)
        plt.xlabel('cosmic time [Gyr]');
    if label != None:
        plt.legend(edgecolor='w')
    plt.ylabel(r'SFR(t) [$M_\odot yr^{-1}$]')
    if logx == True:
        plt.xscale('log')
    if logy == True:
        plt.yscale('log')
    plt.xlim(0,np.amax(timeax))
    tempy = plt.ylim()
    plt.ylim(0,tempy[1])
    # plt.show()
    return fig

def plot_spec(lam, spec, logx = True, logy = True,
xlim = (1e2,1e8),
clip_bottom = True):
    set_plot_style()

    plt.figure(figsize=(12,4))
    plt.plot(lam, spec)
    plt.xlabel(r'$\lambda$')
    plt.ylabel(r'$F_\nu$ [$\mu$Jy]')
    if logx == True:
        plt.xscale('log')
    if logy == True:
        plt.yscale('log')
    plt.xlim(xlim)
    if clip_bottom == True:
        plt.ylim(1e-3,np.amax(spec)*3)
    plt.show()

def plot_filterset(filter_list = 'filter_list_goodss.dat', filt_dir = 'filters/', zval = 1.0, lam_arr = 10**np.linspace(2,8,10000), rest_frame = True):
    set_plot_style()

    filcurves, lam_z, lam_z_lores = make_filvalkit_simple(lam_arr, zval, fkit_name = filter_list, filt_dir = filt_dir)

    plt.figure(figsize=(12,4))
    if rest_frame == True:
        plt.plot(lam_arr*(1+zval), filcurves,'k:');
        plt.xlabel(r'$\lambda$ [$\AA$]');
    else:
        plt.plot(lam_arr, filcurves,'k:');
        plt.xlabel(r'$\lambda$ [$\AA$; obs. frame]');
    plt.xscale('log'); plt.xlim(1e3,2e5)
    plt.ylabel('Filter transmission')
    plt.show()

def quantile_names(N_params):
    return (np.round(np.linspace(0,100,N_params+2)))[1:-1]

def plot_atlas_priors(atlas):
    
    mass_unnormed = np.log10(10**atlas['mstar'] / atlas['norm'])
    sfr_unnormed = np.log10(10**atlas['sfr'] / atlas['norm'])
    ssfr = sfr_unnormed - mass_unnormed
    txs = atlas['sfh_tuple_rec'][0:,3:]

    dust = atlas['dust'].ravel()
    met = atlas['met'].ravel()
    zval = atlas['zval'].ravel()
    quants = np.vstack((mass_unnormed, sfr_unnormed, ssfr, txs.T, met, dust, zval)).T

    txs = ['t'+'%.0f' %i for i in quantile_names(txs.shape[1])]
    pg_labels = ['log M*', 'log SFR', 'log sSFR', 'Z', 'Av', 'z']
    pg_labels[3:3] = txs

    figure = corner.corner(quants,plot_datapoints=False, fill_contours=True,labels=pg_labels,
                                    bins=20, smooth=1.0,
                                    quantiles=(0.16, 0.84), 
                                    levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
                                    label_kwargs={"fontsize": 30})
    figure.subplots_adjust(right=1.5,top=1.5)

    plt.show()
    
    return



def plot_posteriors(chi2_array, norm_fac, sed, atlas, truths = [], **kwargs):
    set_plot_style()

    if len(truths) > 0:
        corner_truths = truths
#         corner_truths[3:(3+int(sed_truths[2]))] = corner_truths[3:(3+int(sed_truths[2]))]/cosmo.age(corner_truths[-1]).value
    #pg_params = np.vstack([pg_theta[0][0,0:], pg_theta[0][1,0:], pg_theta[0][3:,0:], pg_theta[1], pg_theta[2], pg_theta[3]])
    sfrvals = atlas['sfr'].copy()
    sfrvals[sfrvals<-3] = -3
    pg_params = np.vstack([atlas['mstar'],sfrvals,atlas['sfh_tuple'][0:,3:].T,atlas['met'].ravel(),atlas['dust'].ravel(),atlas['zval'].ravel()])
    txs = ['t'+'%.0f' %i for i in quantile_names(pg_params.shape[0]-5)]
    pg_labels = ['log M*', 'log SFR', 'Z', 'Av', 'z']
    pg_labels[2:2] = txs

    corner_params = pg_params.copy()
    corner_params[0,0:] += np.log10(norm_fac)
    corner_params[1,0:] += np.log10(norm_fac)

    if len(truths) > 0:
        figure = corner.corner(corner_params.T, weights = np.exp(-chi2_array/2),
                                labels=pg_labels, truths=corner_truths,
                                plot_datapoints=False, fill_contours=True,
                                bins=20, smooth=1.0,
                                quantiles=(0.16, 0.84), levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
                                label_kwargs={"fontsize": 30}, show_titles=True, **kwargs)
    else:
        figure = corner.corner(corner_params.T, weights = np.exp(-chi2_array/2),
                                labels=pg_labels,
                                plot_datapoints=False, fill_contours=True,
                                bins=20, smooth=1.0,
                                quantiles=(0.16, 0.84), levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)],
                                label_kwargs={"fontsize": 30}, show_titles=True, **kwargs)
    figure.subplots_adjust(right=1.5,top=1.5)
    return figure


def plot_priors(fname, N_pregrid, N_param, dir = 'pregrids/'):
    set_plot_style()

    cat = load_atlas(fname, N_pregrid, N_param, path = dir)
    sfh_tuples = cat['sfh_tuple']
    Av = cat['dust'].ravel()
    Z = cat['met'].ravel()
    z = cat['zval'].ravel()
    seds = cat['sed']
    norm_method = cat['norm_method']
    norm_facs = cat['norm'].ravel()

    pg_theta = [sfh_tuples, Z, Av, z, seds]
    pg_params = np.vstack([pg_theta[0][0,0:], pg_theta[0][1,0:], pg_theta[0][3:,0:], pg_theta[1], pg_theta[2], pg_theta[3]])

    txs = ['t'+'%.0f' %i for i in quantile_names(pg_params.shape[0]-5)]
    pg_labels = ['log M*', 'log SFR', 'Z', 'Av', 'z']
    pg_labels[2:2] = txs

    pg_priors = pg_params.copy()
    pg_priors[0,0:] += np.log10(norm_facs)
    pg_priors[1,0:] += np.log10(norm_facs)
    figure = corner.corner(pg_priors.T,labels=pg_labels, plot_datapoints=False, fill_contours=True,
            bins=50, smooth=1.0, levels=[1 - np.exp(-(1/1)**2/2),1 - np.exp(-(2/1)**2/2)], label_kwargs={"fontsize": 30})
    figure.subplots_adjust(right=1.5,top=1.5)
    plt.show()


def plot_SFH_posterior(chi2_array, norm_fac, sed, atlas, truths = [], plot_ci = True, sfh_threshold = 0.9, **kwargs):
    # to be phased out with a newer function
    set_plot_style()

    #pg_sfhs, pg_Z, pg_Av, pg_z, pg_seds = pg_theta
    pg_sfhs = atlas['sfh_tuple'].T
    pg_z = atlas['zval'].ravel()

    weighted_chi2_indices = np.argsort(np.exp(-chi2_array/2))
    num_sfhs = np.sum(np.exp(-chi2_array[weighted_chi2_indices]/2) > sfh_threshold*(np.exp(-np.amin(chi2_array)/2)))

    temp_sfhs = np.zeros((1000, num_sfhs))
    temp_times = np.zeros((1000, num_sfhs))
    rel_likelihoods = np.zeros((num_sfhs,))
    for i in range(num_sfhs):
        temp_sfh_tuple = pg_sfhs[0:, weighted_chi2_indices[-(i+1)]].copy()
        temp_sfh_tuple[0] = temp_sfh_tuple[0] + np.log10(norm_fac)
        temp_sfh_tuple[1] = temp_sfh_tuple[1] + np.log10(norm_fac)
        temp_sfhs[0:,i], temp_times[0:,i] = tuple_to_sfh(temp_sfh_tuple, zval = pg_z[weighted_chi2_indices[-(i+1)]])
        temp_sfhs[0:,i] = np.flip(correct_for_mass_loss(np.flip(temp_sfhs[0:,i],0), temp_times[0:,i], fsps_time, fsps_massloss),0)

        rel_likelihoods[i] = np.exp(-chi2_array[weighted_chi2_indices[-(i+1)]]/2)

    if plot_ci == False:
        for i in range(num_sfhs):
            if i == 0:
                fig = plot_sfh(temp_times[0:,i], temp_sfhs[0:,i], lookback=True, color='k', alpha=rel_likelihoods[i]**3, **kwargs)
            else:
                plot_sfh(temp_times[0:,i], temp_sfhs[0:,i], lookback=True, fig = fig, color='k', alpha=rel_likelihoods[i]**3, **kwargs)

    if plot_ci == True:
        _, temp_common_time = tuple_to_sfh(temp_sfh_tuple, zval = pg_z[np.argmin(chi2_array)])
        temp_sfhs_splined = np.zeros_like(temp_sfhs)
        fig = plt.figure(figsize=(12,4))
        for i in range(num_sfhs):
            temp_sfhs_splined[0:,i] = np.interp(temp_common_time, temp_times[0:,i], np.flip(temp_sfhs[0:,i],0))
#         plt.fill_between(temp_common_time, np.nanpercentile(temp_sfhs_splined,18,axis=1),
#                          np.nanpercentile(temp_sfhs_splined,84,axis=1), color='k', alpha=0.1)
        qt_array = np.arange(25,76,5)
        for i in range(5):
            if i == 0:
                plt.fill_between(temp_common_time, np.nanpercentile(temp_sfhs_splined,qt_array[i],axis=1),
                                 np.nanpercentile(temp_sfhs_splined,qt_array[-(i+1)],axis=1), color='k', alpha=0.1,
                                 label='25-75 CI')
            else:
                plt.fill_between(temp_common_time, np.nanpercentile(temp_sfhs_splined,qt_array[i],axis=1),
                                 np.nanpercentile(temp_sfhs_splined,qt_array[-(i+1)],axis=1), color='k', alpha=0.1)

        plot_sfh(temp_common_time, np.flip(np.nanmedian(temp_sfhs_splined,axis=1),0),
                       lookback=True, color='k', lw=3, label='median DB-SFH', fig = fig)
#         print(np.nanpercentile(temp_sfhs_splined,49,axis=1))
    if len(truths) == 2:
        plot_sfh(truths[0], truths[1], lookback=True, fig = fig, lw=3,label='true SFH')
        plt.ylim(0,np.amax(truths[1])*1.5)
    plt.legend(edgecolor='w', fontsize=18)
    plt.show()
    return

def plot_SFH_posterior_v2(chi2_array, sed, pg_theta, truths = [], plot_ci = True, sfh_threshold = 0.9, Nbins = 30, max_num = 100, npow = 3, **kwargs):
    set_plot_style()

    pg_sfhs, pg_Z, pg_Av, pg_z, pg_seds = pg_theta
    weighted_chi2_indices = np.argsort(np.exp(-chi2_array/(2*np.amin(chi2_array))))

    num_sfhs = np.sum(np.exp(-chi2_array/2) > sfh_threshold*(np.exp(-np.amin(chi2_array)/2)))
    if num_sfhs > max_num:
        num_sfhs = max_num
        print('truncated to %.0f SFHs to reduce computation time. increase max_num if desired.' %max_num)

    # to-do: add other norm_facs
    norm_fac = np.amax(sed)

    temp_sfh_tuple = pg_sfhs[0:, weighted_chi2_indices[-1]].copy()
    _, temp_common_time = tuple_to_sfh(temp_sfh_tuple, zval = pg_z[np.argmin(chi2_array)])

    temp_sfhs = np.zeros((1000, num_sfhs))
    temp_sfhs_splined = np.zeros_like(temp_sfhs)
    temp_times = np.zeros((1000, num_sfhs))
    rel_likelihoods = np.zeros((num_sfhs,))
    for i in range(num_sfhs):
        temp_sfh_tuple = pg_sfhs[0:, weighted_chi2_indices[-(i+1)]].copy()
        temp_sfh_tuple[0] = temp_sfh_tuple[0] + np.log10(norm_fac)
        temp_sfh_tuple[1] = temp_sfh_tuple[1] + np.log10(norm_fac)
        temp_sfhs[0:,i], temp_times[0:,i] = tuple_to_sfh(temp_sfh_tuple, zval = pg_z[weighted_chi2_indices[-(i+1)]])
        temp_sfhs[0:,i] = np.flip(correct_for_mass_loss(np.flip(temp_sfhs[0:,i],0), temp_times[0:,i], fsps_time, fsps_massloss),0)
        temp_sfhs_splined[0:,i] = np.interp(temp_common_time, temp_times[0:,i], np.flip(temp_sfhs[0:,i],0))
        rel_likelihoods[i] = np.exp(-chi2_array[weighted_chi2_indices[-(i+1)]]/(np.amin(chi2_array)*2))

    sfr_range = np.linspace(0, np.nanpercentile(temp_sfhs_splined,99), Nbins+1)
    sfh_median = np.zeros_like(temp_common_time)
    sfh_posterior = np.zeros((Nbins, len(temp_common_time) ))
    for i in range(len(temp_common_time)):

        a,b = np.histogram(temp_sfhs_splined[i,0:], weights= rel_likelihoods**npow, bins = sfr_range, density=True)
        sfh_posterior[0:,i] = a
        n_c = np.cumsum(a)
        n_c = n_c / np.amax(n_c)
        bin_centers = b[0:-1] + (b[1]-b[0])/2
        sfh_median[i] = bin_centers[np.argmin(np.abs(n_c - 0.5))]

    fig = plt.figure(figsize=(12,4))
    plt.pcolor(temp_common_time, sfr_range[0:-1], sfh_posterior,cmap='magma')
#     clbr = plt.colorbar()
#     clbr.set_label('P(SFR(t))')
    plot_sfh(temp_common_time, sfh_median, lookback=False, fig = fig, lw=3,label='median SFH',color='b')
    if len(truths) == 2:
        plot_sfh(truths[0], truths[1], lookback=True, fig = fig, lw=3,label='true SFH',color='w')
        plt.ylim(0,np.amax(truths[1])*1.5)
    plt.xlim(0,np.amax(temp_common_time))
    plt.xlabel('t [lookback time; Gyr]')
    plt.ylabel('SFR(t)')
    plt.ylim(0,np.amax(sfr_range[0:-1]))
    l = plt.legend(framealpha=0.0)
    for text in l.get_texts():
        text.set_color("white")
    plt.show()

    return


def plot_SFH_posterior_v3(chi2_array, sed, pg_theta, truths = [], plot_ci = True, sfh_threshold = 0.9, Nbins = 30, max_num = 1000, npow = 3, **kwargs):
    set_plot_style()

    pg_sfhs, pg_Z, pg_Av, pg_z, pg_seds = pg_theta
    weighted_chi2_indices = np.argsort(np.exp(-chi2_array/(2*np.amin(chi2_array))))
    Nparam = pg_sfhs.shape[0]-3

    num_sfhs = np.sum(np.exp(-chi2_array/2) > sfh_threshold*(np.exp(-np.amin(chi2_array)/2)))
    if num_sfhs > max_num:
        num_sfhs = max_num
        print('truncated to %.0f SFHs to reduce computation time. increase max_num if desired.' %max_num)

    # to-do: add other norm_facs
    norm_fac = np.amax(sed)

    temp_sfh_tuple = pg_sfhs[0:, weighted_chi2_indices[-1]].copy()
    _, temp_common_time = tuple_to_sfh(temp_sfh_tuple, zval = pg_z[np.argmin(chi2_array)])

    temp_sfhs = np.zeros((1000, num_sfhs))
    temp_sfhs_splined = np.zeros_like(temp_sfhs)
    temp_times = np.zeros((1000, num_sfhs))
    rel_likelihoods = np.zeros((num_sfhs,))
    for i in range(num_sfhs):
        temp_sfh_tuple = pg_sfhs[0:, weighted_chi2_indices[-(i+1)]].copy()
        temp_sfh_tuple[0] = temp_sfh_tuple[0] + np.log10(norm_fac)
        temp_sfh_tuple[1] = temp_sfh_tuple[1] + np.log10(norm_fac)
        temp_sfhs[0:,i], temp_times[0:,i] = tuple_to_sfh(temp_sfh_tuple, zval = pg_z[weighted_chi2_indices[-(i+1)]])
        temp_sfhs[0:,i] = np.flip(correct_for_mass_loss(np.flip(temp_sfhs[0:,i],0), temp_times[0:,i], fsps_time, fsps_massloss),0)
        temp_sfhs_splined[0:,i] = np.interp(temp_common_time, temp_times[0:,i], np.flip(temp_sfhs[0:,i],0))
        rel_likelihoods[i] = np.exp(-chi2_array[weighted_chi2_indices[-(i+1)]]/(np.amin(chi2_array)*2))

    sfr_range = np.linspace(0, np.nanpercentile(temp_sfhs_splined,99), Nbins+1)
    sfh_median = np.zeros_like(temp_common_time)
    sfh_up = np.zeros_like(temp_common_time)
    sfh_dn = np.zeros_like(temp_common_time)
    sfh_posterior = np.zeros((Nbins, len(temp_common_time) ))
    for i in range(len(temp_common_time)):
        a,b = np.histogram(temp_sfhs_splined[i,0:], weights= rel_likelihoods**npow, bins = sfr_range, density=True)
        sfh_posterior[0:,i] = a
        n_c = np.cumsum(a)
        n_c = n_c / np.amax(n_c)
        bin_centers = b[0:-1] + (b[1]-b[0])/2
        sfh_median[i] = bin_centers[np.argmin(np.abs(n_c - 0.5))]
        sfh_up[i] = bin_centers[np.argmin(np.abs(n_c - 0.16))]
        sfh_dn[i] = bin_centers[np.argmin(np.abs(n_c - 0.84))]

    a,b,c = calctimes(temp_common_time, np.flip(sfh_median,0), Nparam+3)
    sfh_median_smooth, time_smooth = tuple_to_sfh(np.hstack([a,b,Nparam+3,c.ravel()]), zval = pg_z[np.argmin(chi2_array)])
    a,b,c = calctimes(temp_common_time, np.flip(sfh_up,0), Nparam+3)
    sfh_up_smooth, time_smooth = tuple_to_sfh(np.hstack([a,b,Nparam+3,c.ravel()]), zval = pg_z[np.argmin(chi2_array)])
    a,b,c = calctimes(temp_common_time, np.flip(sfh_dn,0), Nparam+3)
    sfh_dn_smooth, time_smooth = tuple_to_sfh(np.hstack([a,b,Nparam+3,c.ravel()]), zval = pg_z[np.argmin(chi2_array)])

    fig = plt.figure(figsize=(12,4))
#     plt.pcolor(temp_common_time, sfr_range[0:-1], sfh_posterior,cmap='magma')
#     clbr = plt.colorbar()
#     clbr.set_label('P(SFR(t))')
#     db.plot_sfh(temp_common_time, sfh_median, lookback=False, fig = fig, lw=3,label='median SFH')
    plot_sfh(time_smooth, sfh_median_smooth, lookback=True, color='k', fig = fig, lw=3,label='median SFH')
    plt.fill_between(np.amax(time_smooth) - time_smooth, sfh_dn_smooth, sfh_up_smooth, color='k',alpha=0.1)
    #plt.fill_between(temp_common_time, sfh_dn, sfh_up, color='k',alpha=0.1)

    if len(truths) == 2:
        plot_sfh(truths[0], truths[1], lookback=True, fig = fig, lw=3,label='true SFH')
        plt.ylim(0,np.amax(truths[1])*1.5)
    plt.xlim(0,np.amax(temp_common_time))
    plt.xlabel('t [lookback time; Gyr]')
    plt.ylabel('SFR(t)')
    plt.ylim(0,np.amax(sfr_range[0:-1]))
    l = plt.legend(framealpha=0.0)
#     for text in l.get_texts():
#         text.set_color("white")
    plt.show()

    return



from __future__ import print_function, division, absolute_import

__version__ = "0.1.8"
__bibtex__ = """
@article{iyer2019non,
  title={Non-parametric Star Formation History Reconstruction with Gaussian Processes I: Counting Major Episodes of Star Formation},
  author={Iyer, Kartheik G and Gawiser, Eric and Faber, Sandra M and Ferguson, Henry C and Koekemoer, Anton M and Pacifici, Camilla and Somerville, Rachel},
  journal={arXiv preprint arXiv:1901.02877},
  year={2019}
}
"""  # NOQA

# from . import train_data
# from . import pregrids
# from . import filters

#print('Starting dense_basis. please wait ~ a minute for the FSPS backend to initialize.')

from .priors import *
from .gp_sfh import *
from .sed_fitter import *
from .pre_grid import *
from .tests import *
from .plotter import *
from .mcmc import *
from .parallelization import *
from .basic_fesc import *



import numpy as np
import sys
import warnings
warnings.filterwarnings('ignore')

import george
from george import kernels

from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF,WhiteKernel, ConstantKernel as C, DotProduct, RationalQuadratic, Matern

from scipy.optimize import minimize
from scipy.interpolate import PchipInterpolator, interp1d
import scipy.io as sio

from .priors import *

import pkg_resources

def get_file(folder, filename):
    resource_package = __name__
    resource_path = '/'.join((folder, filename))  # Do not use os.path.join()
    template = pkg_resources.resource_stream(resource_package, resource_path)
    return template

fsps_mlc = sio.loadmat(get_file('train_data','fsps_mass_loss_curve.mat'))
#fsps_mlc = sio.loadmat('dense_basis/train_data/fsps_mass_loss_curve.mat')
fsps_time = fsps_mlc['timeax_fsps'].ravel()
fsps_massloss = fsps_mlc['mass_loss_fsps'].ravel()

# basic SFH tuples
rising_sfh = np.array([10.0,1.0,3,0.5,0.7,0.9])
regular_sfg_sfh = np.array([10.0,0.3,3,0.25,0.5,0.75])
young_quenched_sfh = np.array([10.0,-1.0,3,0.3,0.6,0.8])
old_quenched_sfh = np.array([10.0,-1.0,3,0.1,0.2,0.4])
old_very_quenched_sfh = np.array([10.0,-10.0,3,0.1,0.2,0.4])
double_peaked_SF_sfh = np.array([10.0,0.5,3,0.25,0.4,0.7])
double_peaked_Q_sfh = np.array([10.0,-1.0,3,0.2,0.4,0.8])


def correct_for_mass_loss(sfh, time, mass_loss_curve_time, mass_loss_curve):
    correction_factors = np.interp(time, mass_loss_curve_time, mass_loss_curve)
    return sfh * correction_factors

# functions:
def nll(p, gp, y):
    gp.set_parameter_vector(p)
    ll = gp.log_likelihood(y, quiet=True)
    return -ll if np.isfinite(ll) else 1e25

def grad_nll(p, gp, y):
    gp.set_parameter_vector(p)
    return -gp.grad_log_likelihood(y, quiet=True)

#--------------------------------------------


def gp_interpolator(x,y,res = 1000, Nparam = 3, decouple_sfr = False):

    yerr = np.zeros_like(y)
    yerr[2:(2+Nparam)] = 0.001/np.sqrt(Nparam)
    if len(yerr) > 26:
        yerr[2:(2+Nparam)] = 0.1/np.sqrt(Nparam)
    if decouple_sfr == True:
        yerr[(2+Nparam):] = 0.1
    #if decouple_sfr == True:
    #    yerr[-2:] = 0.1/np.sqrt(Nparam)
    #else:
    #    yerr[-2:] = 0.01/np.sqrt(Nparam)

    #kernel = np.var(yax) * kernels.ExpSquaredKernel(np.median(yax)+np.std(yax))
    #k2 = np.var(yax) * kernels.LinearKernel(np.median(yax),order=1)
    #kernel = np.var(y) * kernels.Matern32Kernel(np.median(y)) #+ k2
    kernel = np.var(y) * (kernels.Matern32Kernel(np.median(y)) + kernels.LinearKernel(np.median(y), order=2))
    gp = george.GP(kernel, solver=george.HODLRSolver)

    #print(xax.shape, yerr.shape)
    gp.compute(x.ravel(), yerr.ravel())

    # optimize kernel parameters
#     p0 = gp.get_parameter_vector()
#     results = minimize(nll, p0, jac=grad_nll, method="L-BFGS-B", args = (gp, y))
#     gp.set_parameter_vector(results.x)

    x_pred = np.linspace(np.amin(x), np.amax(x), res)
    y_pred, pred_var = gp.predict(y.ravel(), x_pred, return_var=True)

    return x_pred, y_pred

def gp_sklearn_interpolator(x,y,res = 1000):

    kernel = DotProduct(10.0, (1e-2,1e2)) *RationalQuadratic(0.1)
    gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=9)
    gp.fit(x.reshape(-1,1),(y-x).reshape(-1,1))

    x_pred = np.linspace(0,1,1000)
    y_pred, sigma = gp.predict(x_pred[:,np.newaxis], return_std=True)
    y_pred = y_pred.ravel() + x_pred

    return x_pred, y_pred

def linear_interpolator(x,y,res = 1000):

    interpolator = interp1d(x,y)
    x_pred = np.linspace(np.amin(x), np.amax(x), res)
    y_pred = interpolator(x_pred)

    return x_pred, y_pred

def Pchip_interpolator(x,y,res = 1000):

    interpolator = PchipInterpolator(x,y)
    x_pred = np.linspace(np.amin(x), np.amax(x), res)
    y_pred = interpolator(x_pred)

    return x_pred, y_pred


def tuple_to_sfh(sfh_tuple, zval, interpolator = 'gp_george', decouple_sfr = False, decouple_sfr_time = 10, sfr_tolerance = 0.05, vb = False,cosmo = cosmo, res = 1000, print_warnings = False):
    # generate an SFH from an input tuple (Mass, SFR, {tx}) at a specified redshift


    Nparam = int(sfh_tuple[2])
    mass_quantiles = np.linspace(0,1,Nparam+2)
    time_quantiles = np.zeros_like(mass_quantiles)
    time_quantiles[-1] = 1
    time_quantiles[1:-1] = sfh_tuple[3:]

    # now add SFR constraints

    # SFR smoothly increasing from 0 at the big bang
    mass_quantiles = np.insert(mass_quantiles,1,[0.00])
    time_quantiles = np.insert(time_quantiles,1,[0.01])

    # SFR constrained to SFR_inst at the time of observation
    #SFH_constraint_percentiles = np.array([0.96,0.97,0.98,0.99])
    SFH_constraint_percentiles = np.array([0.97, 0.98, 0.99])
    for const_vals in SFH_constraint_percentiles:

        delta_mstar = 10**(sfh_tuple[0]) *(1-const_vals)
        delta_t = 1 - delta_mstar/(10**sfh_tuple[1])/(cosmo.age(zval).value*1e9)

        if (delta_t > time_quantiles[-2]) & (delta_t > 0.9):
            mass_quantiles = np.insert(mass_quantiles, -1, [const_vals], )
            time_quantiles = np.insert(time_quantiles, -1, [delta_t],)
        else:
            delta_m = 1 - ((cosmo.age(zval).value*1e9)*(1-const_vals)*(10**sfh_tuple[1]))/(10**sfh_tuple[0])
            time_quantiles = np.insert(time_quantiles, -1, [const_vals])
            mass_quantiles=  np.insert(mass_quantiles, -1, [delta_m])

    if interpolator == 'gp_george':
        time_arr_interp, mass_arr_interp = gp_interpolator(time_quantiles, mass_quantiles, Nparam = int(Nparam), decouple_sfr = decouple_sfr, res=res)
    elif interpolator == 'gp_sklearn':
        time_arr_interp, mass_arr_interp = gp_sklearn_interpolator(time_quantiles, mass_quantiles)
    elif interpolator == 'linear':
        time_arr_interp, mass_arr_interp = linear_interpolator(time_quantiles, mass_quantiles)
    elif interpolator == 'pchip':
        time_arr_interp, mass_arr_interp = Pchip_interpolator(time_quantiles, mass_quantiles)
    else:
        raise Exception('specified interpolator does not exist: {}. \n use one of the following: gp_george, gp_sklearn, linear, and pchip '.format(interpolator))

    sfh_scale = 10**(sfh_tuple[0])/(cosmo.age(zval).value*1e9/1000)
    sfh = np.diff(mass_arr_interp)*sfh_scale
    sfh[sfh<0] = 0
    sfh = np.insert(sfh,0,[0])

    sfr_decouple_time_index = np.argmin(np.abs(time_arr_interp*cosmo.age(zval).value - decouple_sfr_time/1e3))
    if sfr_decouple_time_index == 0:
        sfr_decouple_time_index = 2
    mass_lastbins = np.trapz(x=time_arr_interp[-sfr_decouple_time_index:]*1e9*(cosmo.age(zval).value), y=sfh[-sfr_decouple_time_index:])
    mass_remaining = 10**(sfh_tuple[0]) - mass_lastbins
    if mass_remaining < 0:
        mass_remaining = 0
        if print_warnings == True:
            print('input SFR, M* combination is not physically consistent (log M*: %.2f, log SFR: %.2f.)' %(sfh_tuple[0],sfh_tuple[1]))
    mass_initbins = np.trapz(x=time_arr_interp[0:(1000-sfr_decouple_time_index)]*1e9*(cosmo.age(zval).value), y=sfh[0:(1000-sfr_decouple_time_index)])
    sfh[0:(1000-sfr_decouple_time_index)] = sfh[0:(1000-sfr_decouple_time_index)] * mass_remaining / mass_initbins

    if (np.abs(np.log10(sfh[-1]) - sfh_tuple[1]) > sfr_tolerance) or (decouple_sfr == True):
        sfh[-sfr_decouple_time_index:] = 10**sfh_tuple[1]

#     if decouple_sfr == True:
#         sfr_decouple_time_index = np.argmin(np.abs(time_arr_interp*cosmo.age(zval).value - decouple_sfr_time/1e3))
#         sfh[-sfr_decouple_time_index:] = 10**sfh_tuple[1]

    timeax = time_arr_interp * cosmo.age(zval).value

    if vb == True:
        print('time and mass quantiles:')
        print(time_quantiles, mass_quantiles)
        plt.plot(time_quantiles, mass_quantiles,'--o')
        plt.plot(time_arr_interp, mass_arr_interp)
        plt.axis([0,1,0,1])
        #plt.axis([0.9,1.05,0.9,1.05])
        plt.show()

        print('instantaneous SFR: %.1f' %sfh[-1])
        plt.plot(np.amax(time_arr_interp) - time_arr_interp, sfh)
        #plt.xscale('log')
        plt.show()

    return sfh, timeax

def calctimes(timeax,sfh,nparams):

    massint = np.cumsum(sfh)
    massint_normed = massint/np.amax(massint)
    tx = np.zeros((nparams,))
    for i in range(nparams):
        tx[i] = timeax[np.argmin(np.abs(massint_normed - 1*(i+1)/(nparams+1)))]
        #tx[i] = (np.argmin(np.abs(massint_normed - 1*(i+1)/(nparams+1))))
        #print(1*(i+1)/(nparams+1))

    #mass = np.log10(np.sum(sfh)*1e9)
    mass = np.log10(np.trapz(sfh,timeax*1e9))
    sfr = np.log10(sfh[-1])

    return mass, sfr, tx/np.amax(timeax)

def calctimes_to_tuple(calctimelist):
    nparam = len(calctimelist[2])

    sfhtuple = []
    sfhtuple.append(calctimelist[0])
    sfhtuple.append(calctimelist[1])
    sfhtuple.append(nparam)
    for i in range(nparam):
        sfhtuple.append(calctimelist[2][i])
    return sfhtuple

def scale_t50(t50_val = 1.0, zval = 1.0):
    """
    Change a t50 value from lookback time in Gyr at a given redshift
    to fraction of the age of the universe.

    inputs: t50 [Gyr, lookback time], redshift
    outputs: t50 [fraction of the age of the universe, cosmic time]
    """

    return (1 - t50_val/cosmo.age(zval).value)

def scale_t50_inv(t50_val_frac = 1.0, zval = 1.0):
    """
    Change a t50 value from a fraction of the age of the universe to
    lookback time in Gyr at a given redshift

    inputs: t50 [fraction of the age of the universe, cosmic time], redshift
    outputs: t50 [Gyr, lookback time]
    """

    return (1- t50_val_frac)*cosmo.age(zval).value



from setuptools import setup
import glob
import os

setup(
    name="dense_basis",
    version="0.1.8b",
    author="Kartheik Iyer",
    author_email="kartheik.iyer@dunlap.utoronto.ca",
    url = "https://github.com/kartheikiyer/dense_basis",
    packages=["dense_basis"],
    description="SED fitting with non-parametric star formation histories",
    long_description=open("README.md").read(),
    package_data={"": ["README.md", "LICENSE"], "dense_basis": ["train_data/*.mat", "train_data/alpha_lookup_tables/*.npy", "pregrids/*.mat", "filters/*.dat", "filters/filter_curves/goods_s/*.*","filters/filter_curves/goods_n/*.*", "filters/filter_curves/cosmos/*.*", "filters/filter_curves/egs/*.*", "filters/filter_curves/uds/*.*"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        ],
    install_requires=["george", "corner", "hickle", "schwimmbad"]
)
